/*
 *      aliasdialog.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "aliasdialog.h"
#include "dxutils.h"
#include "config.h"
#include "i18n.h"

AliasEditDialog::AliasEditDialog(const QString& title, QWidget* parent, const QString& alias, const QString& command)
: QDialog(parent)
{
    QVBoxLayout *mainframe = new QVBoxLayout(this);
    QGridLayout *gridframe = new QGridLayout;
    int i = 0;
    gridframe->addWidget(new QLabel(_("Alias:")), i, 0);
    m_alias = new QLineEdit(alias);
    gridframe->addWidget(m_alias, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Command:")), i, 0);
    m_command = new QLineEdit(command);
    gridframe->addWidget(m_command, i, 1);
    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel, Qt::Horizontal);
    connect(buttons, SIGNAL(accepted()), this, SLOT(accept()));
    connect(buttons, SIGNAL(rejected()), this, SLOT(reject()));
    mainframe->addLayout(gridframe);
    mainframe->addWidget(buttons);
    setLayout(mainframe);
    setWindowTitle(title);
}

AliasDialog::AliasDialog(QWidget *parent)
: QDialog(parent)
{
    m_aliases = dxUtils.getAliases();

    QVBoxLayout *mainFrame = new QVBoxLayout(this);

    QStringList headers;
    headers << _("Alias") << _("Command");
    m_table = new QTreeWidget;
    m_table->setHeaderLabels(headers);
    m_table->setSelectionBehavior(QAbstractItemView::SelectRows);
    m_table->setAlternatingRowColors(true);
    m_table->setRootIsDecorated(false);
    m_table->setSelectionMode(QAbstractItemView::SingleSelection);
    QHeaderView *header = m_table->header();
    header->resizeSection(0, 150);
    header->resizeSection(1, 200);
    mainFrame->addWidget(m_table);

    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Cancel);
    m_buttonAdd = new QPushButton(_("&Add"));
    connect(m_buttonAdd, SIGNAL(clicked()), this, SLOT(slotAdd()));
    buttons->addButton(m_buttonAdd, QDialogButtonBox::ActionRole);
    m_buttonModify = new QPushButton(_("&Modify"));
    m_buttonModify->setEnabled(m_aliases.size()?true:false);
    connect(m_buttonModify, SIGNAL(clicked()), this, SLOT(slotModify()));
    buttons->addButton(m_buttonModify, QDialogButtonBox::ActionRole);
    m_buttonDelete = new QPushButton(_("&Delete"));
    m_buttonDelete->setEnabled(m_aliases.size()?true:false);
    connect(m_buttonDelete, SIGNAL(clicked()), this, SLOT(slotDelete()));
    buttons->addButton(m_buttonDelete, QDialogButtonBox::ActionRole);
    m_buttonSaveClose = new QPushButton(_("&Save&&Close"));
    connect(m_buttonSaveClose, SIGNAL(clicked()), this, SLOT(slotSaveClose()));
    buttons->addButton(m_buttonSaveClose, QDialogButtonBox::ActionRole);
    connect(buttons, SIGNAL(rejected()), this, SLOT(reject()));
    mainFrame->addWidget(buttons);
    
    setLayout(mainFrame);
    setWindowTitle(_("Aliases list"));
    
    updateTable();
}

void AliasDialog::slotAdd()
{
    AliasEditDialog *dialog = new AliasEditDialog(_("Alias edit"), this);
    dialog->setModal(true);
    if(dialog->exec() ==  QDialog::Accepted)
    {
        if(dialog->alias().isEmpty() || dialog->command().isEmpty())
            return;
        if(dialog->alias()[0].toLatin1() != '/')
        {
            QMessageBox::information(this, _("Information"), _("Alias command has to start with '/'"));
            return;
        }
        if(dialog->alias().contains(' '))
        {
            QMessageBox::information(this, _("Information"), _("Alias command cann't contain space"));
            return;
        }
        if(commandExist(dialog->alias()))
        {
            QMessageBox::information(this, _("Information"), _("Command '%1' already exists").arg(dialog->alias()));
            return;
        }
        m_aliases.insert(dialog->alias(), dialog->command());
        updateTable();        
    }
}

void AliasDialog::slotModify()
{
    QString oldkey = m_table->currentItem()->text(0);
    AliasEditDialog *dialog = new AliasEditDialog(_("Alias edit"), this, oldkey, m_table->currentItem()->text(1));
    dialog->setModal(true);
    if(dialog->exec() ==  QDialog::Accepted)
    {
        if(dialog->alias().isEmpty() || dialog->command().isEmpty())
            return;
        if(dialog->alias()[0].toLatin1() != '/')
        {
            QMessageBox::information(this, _("Information"), _("Alias command has to start with '/'"));
            return;
        }
        if(dialog->alias().contains(' '))
        {
            QMessageBox::information(this, _("Information"), _("Alias command cann't contain space"));
            return;
        }
        if(commandExist(dialog->alias()) && QString::compare(dialog->alias(), oldkey, Qt::CaseInsensitive))
        {
            QMessageBox::information(this, _("Information"), _("Command '%1' already exists").arg(dialog->alias()));
            return;
        }
        m_aliases.remove(oldkey);
        m_aliases.insert(dialog->alias(), dialog->command());
        updateTable();        
    }
}

void AliasDialog::slotDelete()
{
    if(!m_table->currentItem()) return;
    m_aliases.remove(m_table->currentItem()->text(0));
    if(!m_aliases.size())
    {
        m_buttonModify->setEnabled(false);
        m_buttonDelete->setEnabled(false);
    }
    updateTable();
}

void AliasDialog::slotSaveClose()
{
    dxUtils.setAliases(m_aliases);
    accept();
}

void AliasDialog::updateTable()
{
    m_table->clear();
    QMap<QString,QString>::iterator it;
    for(it=m_aliases.begin(); it!=m_aliases.end(); it++)
    {
        QTreeWidgetItem *item = new QTreeWidgetItem(m_table);
        item->setText(0, it.key());
        item->setText(1, it.value());
    }
    if(m_aliases.size())
    {
        m_table->setCurrentItem(m_table->invisibleRootItem()->child(0));
        m_buttonModify->setEnabled(true);
        m_buttonDelete->setEnabled(true);
    }
}

bool AliasDialog::commandExist(const QString &command)
{
    QMap<QString,QString>::iterator it = m_aliases.find(command);
    if(it != m_aliases.end())
    {
        return true;
    }
    if(dxUtils.isCommand(command[0]=='/' ? command.section('/',1) : command)) return true;
    return false;
}
