/*
 *      serverdialog.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include <QSslSocket>
#include "serverdialog.h"
#include "preferences.h"
#include "config.h"
#include "i18n.h"
#include "dxutils.h"

EditDialog::EditDialog(ServerInfo info, QString title, QWidget *parent) :
    QDialog(parent)
{
    QVBoxLayout *mainframe = new QVBoxLayout(this);
    QGridLayout *gridframe = new QGridLayout;
    int i = 0;
    gridframe->addWidget(new QLabel(_("Network name:")), i, 0);
    m_network = new QLineEdit(info.name);
    gridframe->addWidget(m_network, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Hostname:")), i, 0);
    m_hostname = new QLineEdit(info.hostname);
    gridframe->addWidget(m_hostname, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Port:")), i, 0);
    m_port = new QSpinBox;
    m_port->setRange(1, 65535);
    m_port->setValue(info.port);
    gridframe->addWidget(m_port, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Password:")), i, 0);
    m_password = new QLineEdit(info.passwd);
    m_password->setEchoMode(QLineEdit::PasswordEchoOnEdit);
    gridframe->addWidget(m_password, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Nickname:")), i, 0);
    m_nick = new QLineEdit(info.nick);
    gridframe->addWidget(m_nick, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Realname:")), i, 0);
    m_realname = new QLineEdit(info.realname);
    gridframe->addWidget(m_realname, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Channel(s):")), i, 0);
    m_channels = new QLineEdit;
    m_channels->setText(info.channels.isEmpty()?"#":info.channels);
    m_channels->setToolTip(_("Channels need to be comma separated"));
    gridframe->addWidget(m_channels, i, 1);
    i++;
    gridframe->addWidget(new QLabel(_("Commands on connection:")), i, 0);
    m_commands = new QTextEdit;
    m_commands->setText(info.commands);
    gridframe->addWidget(m_commands, i, 1);
    i++;
    m_ssl = new QCheckBox;
    m_ssl->setChecked(info.useSsl);
#ifndef QT_NO_OPENSSL
    if(QSslSocket::supportsSsl())
    {
        gridframe->addWidget(new QLabel(_("Use SSL:")), i, 0);
        gridframe->addWidget(m_ssl, i, 1);
        i++;
    }
#endif
    gridframe->addWidget(new QLabel(_("Auto connect:")), i, 0);
    m_autoconnect = new QCheckBox;
    m_autoconnect->setChecked(info.autoConnect);
    gridframe->addWidget(m_autoconnect, i, 1);

    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel, Qt::Horizontal);
    connect(buttons, SIGNAL(accepted()), this, SLOT(accept()));
    connect(buttons, SIGNAL(rejected()), this, SLOT(reject()));
    mainframe->addLayout(gridframe);
    mainframe->addWidget(buttons);
    setLayout(mainframe);
    setWindowTitle(title);
}

ServerInfo EditDialog::getInfo() const
{
    ServerInfo info;
    info.name = m_network->text().isEmpty() ? m_hostname->text() : m_network->text();
    info.hostname = m_hostname->text();
    info.port = m_port->value();
    info.nick = m_nick->text();
    info.realname = m_realname->text().isEmpty() ? info.nick : m_realname->text();
    info.passwd = m_password->text();
    info.channels = m_channels->text().length()==1?"":m_channels->text();
    info.commands = m_commands->toPlainText();
    info.autoConnect = m_autoconnect->isChecked();
#ifdef QT_NO_OPENSSL
    info.useSsl = false;
#else
    if(QSslSocket::supportsSsl())
    {
        info.useSsl = m_ssl->isChecked();
    }
    else
    {
        info.useSsl = false;
    }
    return info;
#endif
}

ServerDialog::ServerDialog(QWidget *parent) :
    QDialog(parent)
{
    m_servers = preferences.m_serverList;
    QVBoxLayout *mainFrame = new QVBoxLayout(this);

    QHBoxLayout *serverFrame = new QHBoxLayout;
    m_serverList = new QListWidget;
    connect(m_serverList, SIGNAL(currentRowChanged(int)), this, SLOT(slotChangeServer()));
    connect(m_serverList, SIGNAL(itemDoubleClicked(QListWidgetItem*)), this, SLOT(slotJoin()));
    serverFrame->addWidget(m_serverList);

    QGroupBox *details = new QGroupBox(_("Details"));
    QGridLayout *detailsFrame = new QGridLayout;
    int i = 0;
    QLabel *label = new QLabel(_("Hostname:"));
    m_hostname = new QLabel;
    m_hostname->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_hostname, i, 1);
    i++;
    label = new QLabel(_("Port:"));
    m_port = new QLabel;
    m_port->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_port, i, 1);
    i++;
    label = new QLabel(_("Password:"));
    m_password = new QLabel;
    m_password->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_password, i, 1);
    i++;
    label = new QLabel(_("Nickname:"));
    m_nick = new QLabel;
    m_nick->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_nick, i, 1);
    i++;
    label = new QLabel(_("Realname:"));
    m_realname = new QLabel;
    m_realname->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_realname, i, 1);
    i++;
    label = new QLabel(_("Channel(s):"));
    m_channels = new QLabel;
    m_channels->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_channels, i, 1);
    i++;
    label = new QLabel(_("Commands on connection:"));
    m_commands = new QLabel;
    m_commands->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_commands, i, 1);
    i++;
    label = new QLabel(_("Use SSL:"));
    m_ssl = new QLabel;
    m_ssl->setFrameStyle(QFrame::Sunken|QFrame::Panel);
#ifndef QT_NO_OPENSSL
    if(QSslSocket::supportsSsl())
    {
        detailsFrame->addWidget(label, i, 0);
        detailsFrame->addWidget(m_ssl, i, 1);
        i++;
    }
#endif
    label = new QLabel(_("Auto connect:"));
    m_autoconnect = new QLabel;
    m_autoconnect->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    detailsFrame->addWidget(label, i, 0);
    detailsFrame->addWidget(m_autoconnect, i, 1);
    details->setLayout(detailsFrame);
    serverFrame->addWidget(details);

    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Cancel);
    m_buttonJoin = new QPushButton(_("&Join"));
    m_buttonJoin->setEnabled(m_servers.size()?true:false);
    buttons->addButton(m_buttonJoin, QDialogButtonBox::ActionRole);
    connect(m_buttonJoin, SIGNAL(clicked()), this, SLOT(slotJoin()));
    m_buttonAdd = new QPushButton(_("&Add"));
    buttons->addButton(m_buttonAdd, QDialogButtonBox::ActionRole);
    connect(m_buttonAdd, SIGNAL(clicked()), this, SLOT(slotAdd()));
    m_buttonModify = new QPushButton(_("&Modify"));
    m_buttonModify->setEnabled(m_servers.size()?true:false);
    buttons->addButton(m_buttonModify, QDialogButtonBox::ActionRole);
    connect(m_buttonModify, SIGNAL(clicked()), this, SLOT(slotModify()));
    m_buttonDelete = new QPushButton(_("&Delete"));
    m_buttonDelete->setEnabled(m_servers.size()?true:false);
    buttons->addButton(m_buttonDelete, QDialogButtonBox::ActionRole);
    connect(m_buttonDelete, SIGNAL(clicked()), this, SLOT(slotDelete()));
    m_buttonSaveClose = new QPushButton(_("&Save&&Close"));
    buttons->addButton(m_buttonSaveClose, QDialogButtonBox::ActionRole);
    connect(m_buttonSaveClose, SIGNAL(clicked()), this, SLOT(slotSave()));
    connect(buttons, SIGNAL(rejected()), this, SLOT(reject()));

    mainFrame->addLayout(serverFrame);
    mainFrame->addWidget(buttons);

    setLayout(mainFrame);
    setWindowTitle(_("Servers list"));

    updateList();
    m_serverList->setCurrentRow(0);
    updateDetails();
}

void ServerDialog::slotChangeServer()
{
    updateDetails();
}

void ServerDialog::slotJoin()
{
    preferences.m_serverList = m_servers;
    int current = m_serverList->currentRow();
    emit connectServer(m_servers[current]);
    accept();
}

void ServerDialog::slotAdd()
{
    ServerInfo info;
    info.port = 6667;
    info.useSsl = false;
    info.autoConnect = false;
    EditDialog *dialog = new EditDialog(info, _("Server edit"), this);
    dialog->setModal(true);
    if(dialog->exec() == QDialog::Accepted)
    {
        info = dialog->getInfo();
        if(!info.hostname.isEmpty() && !info.nick.isEmpty() && !hostnameExist(info.hostname, info.port, info.nick))
            m_servers.append(info);
    }
    updateList();
    m_serverList->setCurrentRow(0);
    updateDetails();
}

void ServerDialog::slotDelete()
{
    m_servers.removeAt(m_serverList->currentRow());
    delete m_serverList->currentItem();
}

void ServerDialog::slotModify()
{
    int current = m_serverList->currentRow();
    if(m_servers.size() && current>=0 && current<m_servers.size())
    {
        QString oldhostname = m_servers[current].hostname;
        QString oldnick = m_servers[current].nick;
        int oldport = m_servers[current].port;
        EditDialog *dialog = new EditDialog(m_servers[current], _("Server edit"), this);
        dialog->setModal(true);
        if(dialog->exec() == QDialog::Accepted)
        {
            ServerInfo info = dialog->getInfo();
            if(!info.hostname.isEmpty() && !info.nick.isEmpty() && (!hostnameExist(info.hostname, info.port, info.nick) || (oldhostname==info.hostname && oldport==info.port && oldnick==info.nick)))
                m_servers[current] = info;
        }
    }
    updateList();
    m_serverList->setCurrentRow(0);
    updateDetails();
}

void ServerDialog::slotSave()
{
    preferences.m_serverList = m_servers;
    accept();
}

void ServerDialog::updateList()
{
    m_serverList->clear();
    foreach(ServerInfo info, m_servers)
    {
        QListWidgetItem *item = new QListWidgetItem(info.name);
        m_serverList->addItem(item);
    }
}

void ServerDialog::updateDetails()
{
    int current = m_serverList->currentRow();
    if(m_servers.size() && current>=0 && current<m_servers.size())
    {
        m_hostname->setText(m_servers[current].hostname);
        m_port->setText(QString("%1").arg(m_servers[current].port));
        m_nick->setText(m_servers[current].nick);
        m_realname->setText(m_servers[current].realname);
        m_password->setText(m_servers[current].passwd.isEmpty()?"":"*****");
        m_channels->setText(dxutils::wrapText(m_servers[current].channels, 250, ','));
        m_commands->setText(m_servers[current].commands);
        m_ssl->setText(m_servers[current].useSsl?_("Yes"):_("No"));
        m_autoconnect->setText(m_servers[current].autoConnect?_("Yes"):_("No"));
        m_buttonJoin->setEnabled(true);
        m_buttonModify->setEnabled(true);
        m_buttonDelete->setEnabled(true);
    }
    else
    {
        m_hostname->setText("");
        m_port->setText("");
        m_nick->setText("");
        m_realname->setText("");
        m_password->setText("");
        m_channels->setText("");
        m_commands->setText("");
        m_ssl->setText("");
        m_autoconnect->setText("");
        m_buttonJoin->setEnabled(false);
        m_buttonModify->setEnabled(false);
        m_buttonDelete->setEnabled(false);
    }
}

bool ServerDialog::hostnameExist(const QString &hostname, const int &port, const QString &nick)
{
    for(int i=0; i < m_servers.size(); i++)
    {
        if(m_servers[i].hostname == hostname && m_servers[i].port == port && m_servers[i].nick == nick) return true;
    }
    return false;
}
