/*
 *      ircengine.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include <QDateTime>
#include "ircengine.h"
#include "config.h"
#include "i18n.h"
#include "dxutils.h"
#include "preferences.h"

IrcEngine::IrcEngine(QString channels, QString commands)
    : m_startChannels(dxutils::removeSpaces(channels)), m_startCommands(commands)
{
    m_serverName = "localhost";
    m_networkName = "";
    m_serverPort = 6667;
    m_useSsl = false;
    m_serverPassword = "";
    m_nickName = "xxx";
    m_userName = m_nickName;
    m_realName = m_nickName;
    m_receiveRest = "";
    m_connected = false;
    m_connecting = false;
    m_endmotd = false;
    m_ignoreUserHost = false;
    m_storeChannelList = false;
    m_chanTypes = "#&+!";
    m_myUserHost = "";
    m_ignoreWho = "";
    m_adminPrefix = '!';
    m_ownerPrefix = '*';
    m_opPrefix = '@';
    m_voicePrefix = '+';
    m_halfopPrefix = '%';
    m_attempts = 0;
    m_nickLen = 460;
    m_topicLen = 460;
    m_kickLen = 460;
    m_awayLen = 460;
    m_ping = 0;
    m_socket = new dxSocket;
    m_timerReconnect = new QTimer(this);
    m_timerTryagain = new QTimer(this);
    m_timerPing = new QTimer(this);

    connect(m_timerReconnect, SIGNAL(timeout()), this, SLOT(slotReconnect()));
    connect(m_timerTryagain, SIGNAL(timeout()), this, SLOT(slotTryagain()));
    connect(m_timerPing, SIGNAL(timeout()), this, SLOT(slotPing()));
}

IrcEngine::~IrcEngine()
{
    m_socket->disconnectFrom();
    delete m_socket;
}

void IrcEngine::setUserName(const QString& user)
{
    //needed for freenode
    m_userName = dxutils::removeNonalphanumeric(user);
}

void IrcEngine::slotReconnect()
{
    if(hasReconnectAttempt() && !m_connected)
    {
        if(m_connecting)
        {
            m_timerReconnect->start(1000);
            return;
        }
        clearChannelsCommands(false);
        startConnection();
        m_timerReconnect->start(preferences.m_delayAttempt*1000);
    }
    else
        clearAttempts();
}

//timeout after RPL_TRYAGAIN
void IrcEngine::slotTryagain()
{
    if(!m_ignoreWho.isEmpty() && !m_chanTypes.contains(m_ignoreWho[0]))
    {
        sendWho(m_ignoreWho);
        return;
    }
    if(m_ignoreWhoQueue.size())
    {
        m_ignoreWho = m_ignoreWhoQueue[0];
        m_ignoreWhoQueue.removeFirst();
        sendWho(m_ignoreWho);
    }
    else
        m_ignoreWho = "";
}

//timeout for ping test connection
void IrcEngine::slotPing()
{
    if(m_connecting) return;
    if(m_ping) //we haven't answer for ping timeout
    {
        m_socket->disconnect(this);
        m_socket->disconnectFrom();
        m_connected = false;
        m_ping = 0;
        emit ircError(_("Ping timeout"));
        if(preferences.m_reconnect && hasReconnectAttempt())
        {
            emit ircReconnect(_("Server %1 was disconnected").arg(m_serverName));
            m_timerReconnect->start(preferences.m_delayAttempt*1000);
        }
        else
        {
            m_timerReconnect->stop();
            emit ircDisconnect(_("Server %1 was disconnected").arg(m_serverName));
        }
    }
    else
    {
        sendPingTimeout();
    }
}

void IrcEngine::slotCanRead()
{
    if(m_connected)
    {
        readData();
    }
}

void IrcEngine::slotConnected()
{
    m_connected = true;
    m_connecting = false;
    clearAttempts();
    if(getRemoteIP().isEmpty())
        emit ircConnect(_("Connected to %1 - port %2").arg(m_serverName).arg(m_serverPort));
    else emit ircConnect(_("Connected to %1 (%2) - port %3").arg(m_serverName).arg(getRemoteIP()).arg(m_serverPort));
    if(!m_serverPassword.isEmpty())
        sendLine("PASS "+m_serverPassword+"\r\n");
    sendLine("NICK "+m_nickName+"\r\n");
    sendLine("USER "+m_userName+" 0 * :"+m_realName+"\r\n");
}

void IrcEngine::slotDisconnected()
{
    m_connected = false;
    m_connecting = false;
    m_ping = 0;
    if(preferences.m_reconnect && hasReconnectAttempt())
    {
        emit ircReconnect(_("Server %1 was disconnected").arg(m_serverName));
        m_timerReconnect->start(preferences.m_delayAttempt*1000);
    }
    else
    {
        m_timerReconnect->stop();
        emit ircDisconnect(_("Server %1 was disconnected").arg(m_serverName));
    }
}

void IrcEngine::slotError(QString text)
{
    m_connected = false;
    m_connecting = false;
    m_ping = 0;
    emit ircError(text);
}

void IrcEngine::startConnection()
{
    if(m_connected || m_connecting)
        return;
    connect(m_socket, SIGNAL(socketConnected()), this, SLOT(slotConnected()));
    connect(m_socket, SIGNAL(socketDisconnected()), this, SLOT(slotDisconnected()));
    connect(m_socket, SIGNAL(socketCanRead()), this, SLOT(slotCanRead()));
    connect(m_socket, SIGNAL(socketError(QString)), this, SLOT(slotError(QString)));
    m_connecting = true;
    emit ircConnect(_("Connecting to %1").arg(m_serverName));
    dxutils::debugLine("StartConnection");
    dxutils::debugLine(QString("Attempts on %1-%2-%3: %4").arg(m_serverName).arg(m_serverPort).arg(m_nickName).arg(m_attempts));
    dxutils::debugLine("Connect");
    dxutils::debugLine(QString("startChannels: %1").arg(m_startChannels));
    dxutils::debugLine(QString("startCommands: %1").arg(m_startCommands));
    m_attempts++;
    m_endmotd = false;
    if(m_attempts<2)
    {
        m_socket->setIsSSL(m_useSsl);
        m_socket->setHost(m_serverName, m_serverPort);
    }
    m_socket->connectTo();
}

void IrcEngine::disconnectIrc()
{
    dxutils::debugLine(QString("Quit: %1-%2-%3").arg(m_serverName).arg(m_serverPort).arg(m_nickName));
    if(m_connected) sendLine("QUIT\r\n");
    closeConnection(true);
}

void IrcEngine::disconnectIrc(const QString& reason)
{
    dxutils::debugLine(QString("Quit: %1-%2-%3").arg(m_serverName).arg(m_serverPort).arg(m_nickName));
    if(m_connected) sendLine("QUIT :"+reason+"\r\n");
    closeConnection(true);
}

void IrcEngine::reconnect()
{
    if(m_connecting) return;
    m_connected = false;
    m_ping = 0;
    m_attempts = 0;
    m_ignoreWho = "";
    m_ignoreWhoQueue.clear();
    makeStartChannels();
    m_socket->disconnect(this);
    m_socket->disconnectFrom();
    emit ircReconnectAll(_("Reconnect server %1").arg(m_serverName));
    startConnection();
}

void IrcEngine::closeConnection(bool disableReconnect)
{
    m_ignoreWho = "";
    m_ignoreWhoQueue.clear();
    clearChannelsCommands(disableReconnect);
    if(!m_connected && disableReconnect)
    {
        m_timerReconnect->stop();
        return;
    }
    m_connected = false;
    m_connecting = false;
    m_ping = 0;
    if(preferences.m_reconnect && hasReconnectAttempt() && !disableReconnect)
    {
        emit ircReconnect(_("Server %1 was disconnected").arg(m_serverName));
        m_timerReconnect->start(preferences.m_delayAttempt*1000);
    }
    else
    {
        m_timerReconnect->stop();
        emit ircDisconnect(_("Server %1 was disconnected").arg(m_serverName));
    }
    m_socket->disconnect(this);
    m_socket->disconnectFrom();
}

void IrcEngine::readData()
{
    char buffer[1024];
    int size = 0;
    QString data = m_receiveRest;
    while((size = m_socket->read(buffer, 1023)) > 0)
    {
        buffer[size] = '\0';
        if(dxutils::isUtf8(buffer)) data.append(QString::fromUtf8(buffer));
        else data.append(QString::fromLocal8Bit(buffer));
        while (data.contains('\n'))
        {
            parseLine(data.section('\n',0,0).section('\r',0,0));
            data = data.section('\n',1);
        }
        m_receiveRest = data;
    }
}

void IrcEngine::parseLine(const QString &line)
{
    dxutils::debugLine(QString("<< %1").arg(line));
    QString from, command, params;
    if(line[0] == ':')
    {
        from = dxutils::getParam(line, 1, false).section(':',1);
        command = dxutils::getParam(line, 2, false).toUpper();
        params = dxutils::getParam(line, 3, true);
    }
    else
    {
        from = "";
        command = dxutils::getParam(line, 1, false).toUpper();
        params = dxutils::getParam(line, 2, true);
    }
    if(command.length() == 3)
    {
        int ncommand = command.toInt();
        if((ncommand==376||ncommand==422)&&m_networkName.isEmpty())
            m_networkName = from;
        numeric(ncommand, params);
    }
    else if(command == "PRIVMSG") privmsg(from, params);
    else if(command == "JOIN") join(from, params);
    else if(command == "PART") part(from, params);
    else if(command == "QUIT") quitirc(from, params);
    else if(command == "PING") ping(params);
    else if(command == "PONG") pong(from, params);
    else if(command == "MODE") mode(from, params);
    else if(command == "TOPIC") topic(from, params);
    else if(command == "NOTICE") notice(from, params);
    else if(command == "KICK") kick(from, params);
    else if(command == "NICK") nick(from, params);
    else if(command == "INVITE") invite(from, params);
    else if(command == "ERROR") error(params);
    else unknown(command, params);
    emit ircRaw(line);
}

void IrcEngine::numeric(const int &command, const QString &params)
{
    switch(command) {
        case 5: //RPL_ISUPPORT
        {
            emit ircServerReply(dxutils::getParam(params, 2, true));
            parseRplsupport(params);
        }break;
        case 219: //RPL_ENDOFSTATS
        {
            emit ircServerReply(_("End of STATS report"));
        }break;
        case 242: //RPL_STATSUPTIME
        {
            emit ircServerReply(_("Server Up %1 days %2").arg(dxutils::getParam(params, 4, false)).arg(dxutils::getParam(params, 6, true)));
        }break;
        case 263: //RPL_TRYAGAIN
        {
            //this reply is for 99,9% thru sended WHO
            m_timerTryagain->stop();
            m_timerTryagain->start(60000);
            m_ignoreWho = "";
            dxutils::debugLine("RPL_TRYAGAIN fired!!!");
        }break;
        case 301: //RPL_AWAY
        {
            emit irc301(dxutils::getParam(params, 2, false), dxutils::getParam(params, 3, true).section(':',1));
        }break;
        case 302: //RPL_USERHOST
        {
            if(!m_ignoreUserHost)
                emit ircServerReply(dxutils::getParam(params, 2, true).section(':',1));
            else
            {
                m_ignoreUserHost = false;
                m_myUserHost = m_socket->getIP(dxutils::getParam(params, 2, true).section('@',1));
            }

        }break;
        case 305: //RPL_UNAWAY
        {
            emit irc305(_("You are no longer marked as being away"));
            for(int i=0; i < m_nicks.size(); i++)
            {
                if(m_nicks[i].nick == m_nickName)
                {
                    m_nicks[i].away = false;
                    break;
                }
            }
        }break;
        case 306: //RPL_NOWAWAY
        {
            emit irc306(_("You have been marked as being away"));
            for(int i=0; i < m_nicks.size(); i++)
            {
                if(m_nicks[i].nick == m_nickName)
                {
                    m_nicks[i].away = true;
                    break;
                }
            }
        }break;
        case 311: //RPL_WHOISUSER
        {
            emit ircServerReply(_("Start of WHOIS"));
            emit ircServerReply(dxutils::getParam(params, 2, false)+" ["+dxutils::getParam(params, 3, false)+"@"+dxutils::getParam(params, 4, false)+"]");
            addNick(dxutils::getParam(params, 2, false), dxutils::getParam(params, 3, false), dxutils::getParam(params, 6, true).section(':',1), dxutils::getParam(params, 4, false));
            emit ircServerReply(_("Realname: ")+dxutils::getParam(params, 6, true).section(':',1));
        }break;
        case 312: //RPL_WHOISSERVER
        {
            emit ircServerReply(_("Server: ")+dxutils::getParam(params, 3, false)+" ["+dxutils::getParam(params, 4, true).section(':',1)+"]");
        }break;
        case 313: //RPL_WHOISOPERATOR
        {
            emit ircServerReply(_("%1 is an IRC operator").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 315: //RPL_ENDOFWHO
        {
            bool ignoreEvent = false;
            QString who = dxutils::getParam(params, 2, false);
            if(who.compare(m_ignoreWho, Qt::CaseInsensitive) == 0)
            {
                ignoreEvent = true;
                emit ircAway(who);
                if(m_ignoreWhoQueue.size())
                {
                    m_ignoreWho = m_ignoreWhoQueue[0];
                    m_ignoreWhoQueue.removeFirst();
                    sendWho(m_ignoreWho);
                }
                else
                    m_ignoreWho = "";
                break;
            }
            if(who.compare("*", Qt::CaseInsensitive) == 0)
            {
                ignoreEvent = true;
            }
            if(!ignoreEvent) emit ircServerReply(_("End of WHO list"));
        }break;
        case 317: //RPL_WHOISIDLE
        {
            qlonglong idle = dxutils::getParam(params, 3, false).toLongLong();
            emit ircServerReply(_("Idle: ")+QString("%1:%2:%3").arg(idle/3600).arg((idle/60)%60).arg(idle%60));
            QString datestr = QDateTime::fromTime_t(dxutils::getParam(params, 4, false).toUInt()).toString(Qt::SystemLocaleShortDate);
            emit ircServerReply(_("Signon: ")+datestr);
        }break;
        case 318: //RPL_ENDOFWHOIS
        {
            emit ircServerReply(_("End of WHOIS"));
        }break;
        case 319: //RPL_WHOISCHANNELS
        {
            emit ircServerReply(_("Channels: ")+dxutils::getParam(params, 3, true).section(':',1));
        }break;
        case 320: //RPL_IDENTIFIED
        {
            if(params.contains("is identified to services")) emit ircServerReply(_("%1 : is identified to services").arg(dxutils::getParam(params, 2, false)));
            else if(params.contains("is signed on as account")) emit ircServerReply(_("%1 : is signed on as account %2").arg(dxutils::getParam(params, 2, false)).arg(dxutils::getParam(params, 8, false)));
            else emit ircServerReply(dxutils::getParam(params, 2, true));
        }break;
        case 322: //RPL_LIST
        {
            emit ircServerReply(dxutils::getParam(params, 2, true));
            if(!m_storeChannelList)
            {
                m_listChannels.clear();
                m_storeChannelList = true;
            }
            m_listChannels.append(dxutils::getParam(params, 2, true));
        }break;
        case 323: //RPL_LISTEND
        {
            emit ircServerReply(dxutils::getParam(params, 2, true));
            emit irc323();
            m_storeChannelList = false;
        }break;
        case 324: //RPL_CHANNELMODEIS
        {
            emit ircChmode(dxutils::getParam(params, 2, false), dxutils::getParam(params, 3, false));
        }break;
        case 329: //RPL of channel mode (creation time)
        {
            //now nothing action
        }break;
        case 331: //RPL_NOTOPIC
        {
            emit irc331(dxutils::getParam(params, 2, false), _("No topic is set"));
        }break;
        case 332: //RPL_TOPIC
        {
            emit irc332(dxutils::getParam(params, 2, false), _("Topic for %1 : %2").arg(dxutils::getParam(params, 2, false)).arg(dxutils::getParam(params, 3, true).section(':',1)));
        }break;
        case 333: //RPL_TOPICSETBY
        {
            QString datestr = QDateTime::fromTime_t(dxutils::getParam(params, 4, false).toUInt()).toString(Qt::SystemLocaleShortDate);
            emit irc333(dxutils::getParam(params, 2, false), _("Set %1 %2").arg(dxutils::getParam(params, 3, false).section('!',0,0)).arg(datestr));
        }break;
        case 341: //RPL_INVITING
        {
            emit ircServerReply(_("You invite %1 to %2").arg(dxutils::getParam(params, 2, false)).arg(dxutils::getParam(params, 3, false)));
        }break;
        case 342: //RPL_SUMMONING
        {
            emit ircServerReply(_("%1 :Summoning user to IRC").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 352: //RPL_WHOREPLY
        {
            bool ignoreEvent = false;
            QString nick = dxutils::getParam(params, 6, false);
            QString channel = dxutils::getParam(params, 2, false);
            if(m_ignoreWho.compare(channel, Qt::CaseInsensitive)==0 || m_ignoreWho.compare(nick, Qt::CaseInsensitive)==0)
            {
                //param:xxx #test ~dvx localhost dvx.irc.dvx dvx H :0 dvx
                addNick(nick, dxutils::getParam(params, 3, false), dxutils::getParam(params, 9, true), dxutils::getParam(params, 4, false), dxutils::getParam(params, 7, false)[0] == 'H' ? false : true);
                ignoreEvent = true;
                break;
            }
            if(!ignoreEvent) emit ircServerReply(dxutils::getParam(params, 2, true));
        }break;
        case 353: //RPL_NAMREPLY
        {
            bool list = false;
            QString channel = dxutils::getParam(params, 3, false);
            for(int i=0; i<m_namesQueue.size(); i++)
            {
                if(channel==m_namesQueue[i])
                {
                    list = true;
                    break;
                }
            }
            QString nicks = dxutils::getParam(params, 4, true);
            //checking ':' is for servers, which don't use RFC right :)
            emit irc353(channel, nicks[0]==':' ? nicks.section(':',1) : nicks, list);
        }break;
        case 366: //RPL_ENDOFNAMES
        {
            bool list = false;
            QString channel = dxutils::getParam(params, 2, false);
            for(int i=0; i<m_namesQueue.size(); i++)
            {
                if(channel==m_namesQueue[i])
                {
                    list = true;
                    m_namesQueue.removeAt(i);
                    break;
                }
            }
            emit irc366(channel, dxutils::getParam(params, 3, true).section(':',1), list);
        }break;
        case 367: //RPL_BANLIST
        {
            QString datestr = dxutils::getParam(params, 5, true);
            if(datestr.isEmpty()) emit ircServerReply(dxutils::getParam(params, 2, false)+" "+dxutils::getParam(params, 3, false));
            else
            {
                QDateTime time = QDateTime::fromTime_t(datestr.toUInt());
                datestr = time.toString(Qt::SystemLocaleShortDate);
                emit ircServerReply(dxutils::getParam(params, 2, false)+" "+dxutils::getParam(params, 3, false)+" "+dxutils::getParam(params, 4, false)+" "+datestr);
            }
        }break;
        case 369: //RPL_ENDOFWHOWAS
        {
            emit ircServerReply(_("End of WHOWAS"));
        }break;
        case 372: //RPL_MOTD
        {
            emit irc372(dxutils::getParam(params, 2, true).section(':',1));
        }break;
        case 374: //RPL_ENDOFINFO
        {
            emit ircServerReply(_("End of INFO list"));
        }break;
        case 376: //RPL_ENDOFMOTD
        {
            emit ircServerReply(_("End of /MOTD command"));
            emit ircEndMotd();
            if(m_nickName != dxutils::getParam(params, 1, false)) m_nickName = dxutils::getParam(params, 1, false);
            m_endmotd = true;
            if(!m_startChannels.isEmpty())
            {
                sendJoin(m_startChannels);
                m_startChannels.clear();
            }
            if(!m_startCommands.isEmpty())
            {
                sendCommands();
            }
            sendPingTimeout();
            sendLine("USERHOST "+m_nickName+" \r\n");
            m_ignoreUserHost = true;
        }break;
        case 381: //RPL_YOUREOPER
        {
            emit ircServerReply(_("You are now an IRC operator"));
        }break;
        case 401: //ERR_NOSUCHNICK
        {
            emit irc401(dxutils::getParam(params, 2, false));
            emit ircServerError(_("%1 :No such nick/channel").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 402: //ERR_NOSUCHSERVER
        {
            emit ircServerError(_("%1 :No such server").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 403: //ERR_NOSUCHCHANNEL
        {
            emit ircServerError(_("%1 :No such channel").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 404: //ERR_CANNOTSENDTOCHAN
        {
            emit ircServerError(_("%1 :Cannot send to channel").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 405: //ERR_TOOMANYCHANNELS
        {
            emit ircServerError(_("%1 :You have joined too many channels").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 406: //ERR_WASNOSUCHNICK
        {
            emit ircServerError(_("%1 :There was no such nickname").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 407: //ERR_TOOMANYTARGETS
        {
            emit ircServerError(_("%1 :Too many targets").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 408: //ERR_NOSUCHSERVICE
        {
            emit ircServerError(_("%1 :No such service").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 409: //ERR_NOORIGIN
        {
            emit ircServerError(_("No origin specified"));
        }break;
        case 411: //ERR_NORECIPIENT
        {
            emit ircServerError(_("No recipient given"));
        }break;
        case 412: //ERR_NOTEXTTOSEND
        {
            emit ircServerError(_("No text to send"));
        }break;
        case 413: //ERR_NOTOPLEVEL
        {
            emit ircServerError(_("%1 :No toplevel domain specified").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 414: //ERR_WILDTOPLEVEL
        {
            emit ircServerError(_("%1 :Wildcard in toplevel domain").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 415: //ERR_BADMASK
        {
            emit ircServerError(_("%1 :Bad Server/host mask").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 421: //ERR_UNKNOWNCOMMAND
        {
            emit ircServerError(_("%1 :Unknown command").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 422: //ERR_NOMOTD
        {
            emit ircServerError(_("MOTD File is missing"));
            emit ircEndMotd();
            if(m_nickName != dxutils::getParam(params, 1, false)) m_nickName = dxutils::getParam(params, 1, false);
            m_endmotd = true;
            if(!m_startChannels.isEmpty())
            {
                sendJoin(m_startChannels);
                m_startChannels.clear();
            }
            if(!m_startCommands.isEmpty())
            {
                sendCommands();
            }
            sendPingTimeout();
            sendLine("USERHOST "+m_nickName+" \r\n");
            m_ignoreUserHost = true;
        }break;
        case 423: //ERR_NOADMININFO
        {
            emit ircServerError(_("%1 :No administrative info available").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 424: //ERR_FILEERROR
        {
            emit ircServerError(_("File error doing"));
        }break;
        case 431: //ERR_NONICKNAMEGIVEN
        {
            emit ircServerError(_("No nickname given"));
        }break;
        case 432: //ERR_ERRONEUSNICKNAME
        {
            emit ircServerError(_("%1 :Erroneous nickname").arg(dxutils::getParam(params, 2, false)));
            if(m_endmotd) emit ircServerReply(_("You still have nick: %1").arg(m_nickName));
            else
            {
                m_nickName = "_xxx_";
                sendNick(m_nickName);
            }
        }break;
        case 433: //ERR_NICKNAMEINUSE
        {
            emit ircServerError(_("%1 :Nickname is already in use").arg(dxutils::getParam(params, 2, false)));
            if(m_endmotd) emit ircServerReply(_("You still have nick: %1").arg(m_nickName));
            else
            {
                m_nickName += "_";
                sendNick(m_nickName);
            }
        }break;
        case 436: //ERR_NICKCOLLISION
        {
            emit ircServerError(_("%1 :Nickname collision").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 437: //ERR_UNAVAILRESOURCE
        {
            emit ircServerError(_("%1 :Nick/channel is temporarily unavailable").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 441: //ERR_USERNOTINCHANNEL
        {
            emit ircServerError(_("%1 %2 :They aren't on that channel").arg(dxutils::getParam(params, 2, false)).arg(dxutils::getParam(params, 3, false)));
        }break;
        case 442: //ERR_NOTONCHANNEL
        {
            emit ircServerError(_("%1 :You're not on that channel").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 443: //ERR_USERONCHANNEL
        {
            emit ircServerError(_("%1 %2 :is already on channel").arg(dxutils::getParam(params, 2, false)).arg(dxutils::getParam(params, 3, false)));
        }break;
        case 444: //ERR_NOLOGIN
        {
            emit ircServerError(_("%1 :User not logged in").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 445: //ERR_SUMMONDISABLED
        {
            emit ircServerError(_("SUMMON has been disabled"));
        }break;
        case 446: //ERR_USERSDISABLED
        {
            emit ircServerError(_("USERS has been disabled"));
        }break;
        case 451: //ERR_NOTREGISTERED
        {
            emit ircServerError(_("You have not registered"));
        }break;
        case 461: //ERR_NEEDMOREPARAMS
        {
            emit ircServerError(_("%1 :Not enough parameters").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 462: //ERR_ALREADYREGISTRED
        {
            emit ircServerError(_("Unauthorized command (already registered)"));
        }break;
        case 463: //ERR_NOPERMFORHOST
        {
            emit ircServerError(_("Your host isn't among the privileged"));
        }break;
        case 464: //ERR_PASSWDMISMATCH
        {
            emit ircServerError(_("Password incorrect"));
        }break;
        case 465: //ERR_YOUREBANNEDCREEP
        {
            emit ircServerError(_("You are banned from this server"));
        }break;
        case 466: //ERR_YOUWILLBEBANNED
        {
            emit ircServerError(_("You will be banned from this server"));
        }break;
        case 467: //ERR_KEYSET
        {
            emit ircServerError(_("%1 :Channel key already set").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 471: //ERR_CHANNELISFULL
        {
            emit ircServerError(_("%1 :Cannot join channel (+l)").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 472: //ERR_UNKNOWNMODE
        {
            emit ircServerError(_("%1 :is unknown mode char to me for channel").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 473: //ERR_INVITEONLYCHAN
        {
            emit ircServerError(_("%1 :Cannot join channel (+i)").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 474: //ERR_BANNEDFROMCHAN
        {
            emit ircServerError(_("%1 :Cannot join channel (+b)").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 475: //ERR_BADCHANNELKEY
        {
            emit ircServerError(_("%1 :Cannot join channel (+k)").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 476: //ERR_BADCHANMASK
        {
            if(dxutils::getParam(params, 3, true).contains("Bad Channel Mask")) emit ircServerError(_("%1 :Bad Channel Mask").arg(dxutils::getParam(params, 2, false)));
            else emit ircServerError(params);
        }break;
        case 477: //ERR_NOCHANMODES
        {
            if(dxutils::getParam(params, 3, true).contains("Channel doesn't support modes")) emit ircServerError(_("%1 :Channel doesn't support modes").arg(dxutils::getParam(params, 2, false)));
            else emit ircServerError(params);
        }break;
        case 478: //ERR_BANLISTFULL
        {
            if(dxutils::getParam(params, 3, true).contains("Channel list is full")) emit ircServerError(_("%1 %2 :Channel list is full").arg(dxutils::getParam(params, 2, false)).arg(dxutils::getParam(params, 3, false)));
            else emit ircServerError(params);
        }break;
        case 481: //ERR_NOPRIVILEGES
        {
            emit ircServerError(_("Permission Denied - You're not an IRC operator"));
        }break;
        case 482: //ERR_CHANOPRIVSNEEDED
        {
            emit ircServerError(_("%1 :You're not channel operator").arg(dxutils::getParam(params, 2, false)));
        }break;
        case 483: //ERR_CANTKILLSERVER
        {
            emit ircServerError(_("You can't kill a server!"));
        }break;
        case 484: //ERR_RESTRICTED
        {
            emit ircServerError(_("Your connection is restricted!"));
        }break;
        case 485: //ERR_UNIQOPPRIVSNEEDED
        {
            emit ircServerError(_("You're not the original channel operator"));
        }break;
        case 491: //ERR_NOOPERHOST
        {
            emit ircServerError(_("No O-lines for your host"));
        }break;
        case 501: //ERR_UMODEUNKNOWNFLAG
        {
            emit ircServerError(_("Unknown MODE flag"));
        }break;
        case 502: //ERR_USERSDONTMATCH
        {
            emit ircServerError(_("Cannot change mode for other users"));
        }break;
        default:
        {
            if(dxutils::getParam(params, 2, true)[0] == ':') emit ircServerReply(dxutils::getParam(params, 2, true).section(':',1));
            else emit ircServerReply(dxutils::getParam(params, 2, true));
        }
    }
}

/*
 * Parse 005 - RPL_ISUPPORT
 * http://www.irc.org/tech_docs/005.html
 */
void IrcEngine::parseRplsupport(QString text)
{
    // nickLen, topicLen, kickLen, awayLen
    if(text.right(1) != " ") text.append(" ");
    QStringList list = text.split(' ',QString::SkipEmptyParts);
    for(int i=0; i<list.size(); i++)
    {
        QString parameter = list[i].section('=',0,0);
        QString value = list[i].section('=',1);
        if(value.isEmpty())
            continue;
        if(!parameter.compare("PREFIX", Qt::CaseInsensitive))
        {
            QString modes = value.section('(',1).section(')',0,0);
            QString chars = value.section(')',1);
            if(modes.length() != chars.length())
                continue;
            for(int j=0; j<modes.length(); j++)
            {
                switch(modes[j].toLatin1()) {
                    case 'a':
                        m_adminPrefix = chars[j].toLatin1();
                        break;
                    case 'q':
                        m_ownerPrefix = chars[j].toLatin1();
                        break;
                    case 'o':
                        m_opPrefix = chars[j].toLatin1();
                        break;
                    case 'v':
                        m_voicePrefix = chars[j].toLatin1();
                        break;
                    case 'h':
                        m_halfopPrefix = chars[j].toLatin1();
                        break;
                }
            }
            continue;
        }
        if(!parameter.compare("CHANTYPES", Qt::CaseInsensitive))
        {
            m_chanTypes = value;
            continue;
        }
        if(!parameter.compare("NETWORK", Qt::CaseInsensitive))
        {
            m_networkName = value;
            continue;
        }
        if(!parameter.compare("NICKLEN", Qt::CaseInsensitive))
        {
            m_nickLen = value.toInt();
            continue;
        }
        if(!parameter.compare("TOPICLEN", Qt::CaseInsensitive))
        {
            m_topicLen = value.toInt();
            continue;
        }
        if(!parameter.compare("KICKLEN", Qt::CaseSensitive))
        {
            m_kickLen = value.toInt();
            continue;
        }
        if(!parameter.compare("AWAYLEN", Qt::CaseInsensitive))
        {
            m_awayLen = value.toInt();
            continue;
        }
    }
    dxutils::debugLine(QString("adminPrefix=%1;ownerPrefix=%2;opPrefix=%3;voicePrefix=%4;halfopPrefix=%5")
                        .arg(m_adminPrefix).arg(m_ownerPrefix).arg(m_opPrefix).arg(m_voicePrefix).arg(m_halfopPrefix));
    dxutils::debugLine(QString("chantypes=%1").arg(m_chanTypes));
    dxutils::debugLine(QString("nickLen=%1").arg(m_nickLen));
    dxutils::debugLine(QString("topicLen=%1").arg(m_topicLen));
    dxutils::debugLine(QString("kickLen=%1").arg(m_kickLen));
    dxutils::debugLine(QString("awayLen=%1").arg(m_awayLen));
}

void IrcEngine::privmsg(const QString &from, const QString &params)
{
    QString nick = from.section('!',0,0);
    QString to = dxutils::getParam(params, 1, false);
    QString msg = dxutils::getParam(params, 2, true).section(':',1);
    if(msg[0] == '\001') ctcp(from, params);
    else
    {
        if(m_nickName == to && !isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to)) emit ircQuery(nick);
        if(!isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to)) emit ircPrivmsg(nick, to, msg);
    }
}

void IrcEngine::ctcp(const QString &from, const QString &params)
{
    QString nick = from.section('!',0,0);
    QString to = dxutils::getParam(params, 1, false);
    QString msg = dxutils::getParam(params, 2, true).section(':',1).section('\001',1).section('\001',0,0);
    QString ctcpCommand = msg.section(' ',0,0).toUpper();
    QString ctcpRest = msg.section(' ',1);
    if(ctcpCommand == "VERSION")
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpRequest(nick, ctcpCommand);
        sendVersion(nick);
    }
    else if(ctcpCommand == "ACTION")
    {
        if(m_nickName == to && !isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to)) emit ircQuery(nick);
        if(!isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to)) emit ircAction(nick, to, ctcpRest);
    }
    else if(ctcpCommand == "PING")
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpRequest(nick, ctcpCommand);
        sendCtcpNotice(nick, ctcpCommand+" "+ctcpRest);
    }
    else if(ctcpCommand == "DCC")
    {
        QString dccCommand = dxutils::getParam(ctcpRest, 1, false).toUpper();
        if(dccCommand == "CHAT")
        {
            if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
                return;
            emit ircDccChat(nick, m_socket->binaryIPToString(dxutils::getParam(ctcpRest, 3, false)), dxutils::getParam(ctcpRest, 4, false).toInt());
        }
        if(dccCommand == "SEND")
        {
            if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
                return;
            if(dxutils::getParam(ctcpRest, 5, false) != dxutils::getParam(ctcpRest, 6, false)) //passive send
            {
                if(dxutils::getParam(ctcpRest, 4, false).toInt())
                    emit ircDccMyToken(m_socket->binaryIPToString(dxutils::getParam(ctcpRest, 3, false)), dxutils::getParam(ctcpRest, 4, false).toInt(), dxutils::getParam(ctcpRest, 6, false).toInt());
                else
                    emit ircDccToken(nick, dxutils::getParam(ctcpRest, 2, false), dxutils::getParam(ctcpRest, 5, false), dxutils::getParam(ctcpRest, 6, false).toInt());
            }
            else
                emit ircDccIn(nick, m_socket->binaryIPToString(dxutils::getParam(ctcpRest, 3, false)), dxutils::getParam(ctcpRest, 4, false), dxutils::getParam(ctcpRest, 2, false), dxutils::getParam(ctcpRest, 5, false));
        }
        if(dccCommand == "RESUME")
        {
            // 2-filename,3-port,4-position,5-token
            if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
                return;
            if(dxutils::getParam(ctcpRest, 4, false) != dxutils::getParam(ctcpRest, 5, false)) //passive send
                emit ircDccPresume(dxutils::getParam(ctcpRest, 5, false).toInt(), dxutils::getParam(ctcpRest, 4, false));
            else
                emit ircDccResume(nick, dxutils::getParam(ctcpRest, 2, false), dxutils::getParam(ctcpRest, 3, false), dxutils::getParam(ctcpRest, 4, false));
        }
        if(dccCommand == "ACCEPT")
        {
            // 2-filename,3-port,4-position,5-token
            if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
                return;
            if(dxutils::getParam(ctcpRest, 4, false) != dxutils::getParam(ctcpRest, 5, false)) //passive send
                emit ircDccPaccept(dxutils::getParam(ctcpRest, 5, false).toInt(), dxutils::getParam(ctcpRest, 4, false));
            else
                emit ircDccAccept(nick, dxutils::getParam(ctcpRest, 2, false), dxutils::getParam(ctcpRest, 3, false), dxutils::getParam(ctcpRest, 4, false));
        }
    }
    else if(ctcpCommand == "USERINFO")
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpRequest(nick, ctcpCommand);
        sendCtcpNotice(nick, ctcpCommand+" "+m_realName);
    }
    else if(ctcpCommand == "CLIENTINFO")
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpRequest(nick, ctcpCommand);
        sendCtcpNotice(nick, "CLIENTINFO ACTION BOATS CLIENTINFO PING TIME USERINFO VERSION");
    }
    else if(ctcpCommand == "TIME")
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpRequest(nick, ctcpCommand);
        sendCtcpNotice(nick, ctcpCommand+" "+QDateTime::currentDateTime().toString(Qt::SystemLocaleShortDate));
    }
    else if(ctcpCommand == "BOATS")
    {
        QString boatsCommand = dxutils::getParam(ctcpRest, 1, false).toUpper();
        if(boatsCommand == "INVITE")
        {
            emit ircBoatsInvite(nick);
        }
        if(boatsCommand == "ACCEPT")
        {
            emit ircBoatsAccept(nick);
        }
        if(boatsCommand == "DECLINE")
        {
            emit ircBoatsDecline(nick);
        }
        if(boatsCommand == "NEWGAME")
        {
            emit ircBoatsNewGame(nick);
        }
        if(boatsCommand == "READY")
        {
            emit ircBoatsReady(nick);
        }
        if(boatsCommand == "LEFT")
        {
            emit ircBoatsLeft(nick);
        }
        if(boatsCommand == "DESTROYED")
        {
            emit ircBoatsDestroyed(nick);
        }
        if(boatsCommand == "WIN")
        {
            emit ircBoatsWin(nick);
        }
        /*
         * dxutils::getParam(ctcpRest, 1, false) - position
         * dxutils::getParam(ctcpRest, 2, true) - status: FIRE, FAIL, HIT
         *
         */
        emit ircBoats(nick, dxutils::getParam(ctcpRest, 1, false).toUpper(), dxutils::getParam(ctcpRest, 2, true).toUpper());
    }
    else
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpRequest(nick, ctcpCommand);
    }
}

void IrcEngine::join(const QString &from, const QString &params)
{
    QString channel = dxutils::getParam(params, 1, false);
    if(channel[0] == ':') channel = channel.section(':',1);
    QString nick = from.section('!',0,0);
    if(m_nickName == nick)
    {
        m_channels.append(channel);
        emit ircNewChannel(channel[0] == '&' ? "&"+channel : channel);
        addNick(m_nickName, m_userName, m_realName, from.section('@',1));
        return;
    }
    else
    {
        emit ircJoin(nick, channel);
        if(m_ignoreWho==nick)
        {
            return;
        }
        for(int i=0; i < m_nicks.size(); i++)
        {
            if(m_nicks[i].nick==nick)
                return;
        }
        addIgnoreWho(nick); //for update NickInfo
        addNick(nick, from.section('!',1).section('@',0,0), "", from.section('@',1), false);
    }
}

void IrcEngine::quitirc(const QString &from, const QString &params)
{
    QString reason = dxutils::getParam(params, 1, true).section(':',1);
    QString nick = from.section('!',0,0);
    if(m_nickName == nick) m_channels.clear();
    emit ircQuit(nick, reason);
    removeNick(nick);
}

void IrcEngine::part(const QString &from, const QString &params)
{
    QString channel = dxutils::getParam(params, 1, false);
    QString reason = dxutils::getParam(params, 1, true).section(':',1);
    QString nick = from.section('!',0,0);
    if(m_nickName == nick) m_channels.removeAll(channel);
    emit ircPart(nick, channel, reason);
}

void IrcEngine::ping(const QString &params)
{
    sendLine("PONG "+params+"\r\n");
}

void IrcEngine::pong(const QString &from, const QString &params)
{
    QString arg = dxutils::getParam(params, 2, true);
    qlonglong pong = arg[0]==':'?arg.section(':',1).toLongLong():arg.toLongLong();
    if(m_ping && m_ping==pong)
        m_ping = 0;
    else
        emit ircServerReply(_("PONG from %1 %2").arg(from).arg(pong));
}

void IrcEngine::notice(const QString &from, const QString &params)
{
    QString to = dxutils::getParam(params, 1, false);
    QString nick = from.section('!',0,0);
    QString msg = dxutils::getParam(params, 2, true).section(':',1);
    if(msg[0] == '\001')
    {
        if(isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
            return;
        emit ircCtcpReply(nick, msg.section('\001',1).section('\001',0,0));
    }
    else
    {
        if(from.isEmpty()) //public notice
        {
            emit ircNotice(to, msg);
        }
        else
        {
            if(!isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), to))
                emit ircChnotice(nick, to, msg);
        }
    }
}

void IrcEngine::nick(const QString &from, const QString &params)
{
    QString nick = from.section('!',0,0);
    QString newnick = dxutils::getParam(params, 1, false);
    if(newnick[0] == ':') newnick = newnick.section(':',1);
    if(m_nickName == nick) m_nickName = newnick;
    for(int i = 0; i < m_nicks.size(); i++)
    {
        if(nick == m_nicks[i].nick)
        {
            m_nicks[i].nick = newnick;
            break;
        }
    }
    emit ircNick(nick, newnick);
}

void IrcEngine::topic(const QString &from, const QString &params)
{
    QString nick = from.section('!',0,0);
    QString channel = dxutils::getParam(params, 1, false);
    QString topic = dxutils::getParam(params, 2, true).section(':',1);
    emit ircTopic(nick, channel, topic);
}

void IrcEngine::invite(const QString &from, const QString &params)
{
    QString nick = from.section('!',0,0);
    QString channel = dxutils::getParam(params, 2, false);
    if(channel[0] == ':') channel = channel.section(':',1);
    if(!isUserIgnored(nick, from.section('!',1).section('@',0,0), from.section('@',1), channel)) emit ircInvite(nick, channel);
}

void IrcEngine::kick(const QString &from, const QString &params)
{
    QString nick = from.section('!',0,0);
    QString to = dxutils::getParam(params, 2, false);
    QString channel = dxutils::getParam(params, 1, false);
    QString reason = dxutils::getParam(params, 3, true).section(':',1);
    emit ircKick(nick, to, channel, reason);
}

void IrcEngine::mode(const QString &from, const QString &params)
{
    if(params.contains(':')) emit ircMode(dxutils::getParam(params, 2, true).section(':',1), dxutils::getParam(params, 1, false));
    else //channel mode
    {
        QString moderator = from.section('!',0,0);
        QString channel = dxutils::getParam(params, 1, false);
        QString modes = dxutils::getParam(params, 2, false);
        QString args = dxutils::getParam(params, 3, true);
        emit ircUmode(moderator, channel, modes, args);
        //hack for &bitlbee
        if(channel=="&bitlbee" && modes.count()>1 && modes[1]=='v') addIgnoreWho(args);
    }
}

void IrcEngine::error(const QString &params)
{
    emit ircServerError(params[0] == ':' ? params.section(':',1) : params);
}

void IrcEngine::unknown(const QString &command, const QString &params)
{
    emit ircUnknown(command, params);
}

void IrcEngine::addIgnoreWho(const QString& who)
{
    if(m_ignoreWhoQueue.size())
    {
        for(int i=0; i<m_ignoreWhoQueue.size(); i++)
        {
            if(who==m_ignoreWhoQueue[i])
                return;
        }
        m_ignoreWhoQueue.append(who);
    }
    else
    {
        if(m_ignoreWho.isEmpty())
        {
            m_ignoreWho = who;
            if(!m_timerTryagain->isActive()) sendWho(who);
            else m_ignoreWhoQueue.append(who);
        }
        else
            m_ignoreWhoQueue.append(who);
    }
}

bool IrcEngine::sendAdmin(const QString& params)
{
    return sendLine("ADMIN "+params+"\r\n");
}

bool IrcEngine::sendAway(const QString& params)
{
    return sendLine("AWAY :"+params+"\r\n");
}

bool IrcEngine::sendBanlist(const QString& channel)
{
    return sendLine("MODE "+channel+" +b\r\n");
}

bool IrcEngine::sendCtcp(const QString& to, const QString& params)
{
    return sendLine("PRIVMSG "+to+" :\001"+params+"\001\r\n");
}

bool IrcEngine::sendCtcpNotice(const QString& to, const QString& params)
{
    return sendLine("NOTICE "+to+" :\001"+params+"\001\r\n");
}

bool IrcEngine::sendMode(const QString& params)
{
    return sendLine("MODE "+params+"\r\n");
}

bool IrcEngine::sendInvite(const QString& to, const QString& params)
{
    return sendLine("INVITE "+to+" "+params+"\r\n");
}

bool IrcEngine::sendJoin(const QString& chan)
{
    return sendLine("JOIN "+chan+"\r\n");
}

bool IrcEngine::sendKick(const QString& chan, const QString& nick, const QString& reason)
{
    return sendLine("KICK "+chan+" "+nick+" :"+reason+"\r\n");
}

bool IrcEngine::sendKill(const QString& nick, const QString& reason)
{
    return sendLine("KILL "+nick+" :"+reason+"\r\n");
}

bool IrcEngine::sendList(const QString& params)
{
    return sendLine("LIST "+params+"\r\n");
}

bool IrcEngine::sendMe(const QString& to, const QString& message)
{
    return sendLine("PRIVMSG "+to+" :\001ACTION "+message+"\001\r\n");
}

bool IrcEngine::sendMsg(const QString& to, const QString& message)
{
    return sendLine("PRIVMSG "+to+" :"+message+"\r\n");
}

bool IrcEngine::sendNames(const QString& channel, bool list)
{
    if(list) m_namesQueue.append(channel);
    return sendLine("NAMES "+channel+"\r\n");
}

bool IrcEngine::sendNick(const QString& nick)
{
    return sendLine("NICK "+nick+"\r\n");
}

bool IrcEngine::sendNotice(const QString& to, const QString& message)
{
    return sendLine("NOTICE "+to+" :"+message+"\r\n");
}

bool IrcEngine::sendOper(const QString& login, const QString& password)
{
    return sendLine("OPER "+login+" "+password+"\r\n");
}

bool IrcEngine::sendPart(const QString& chan)
{
    return sendLine("PART "+chan+"\r\n");
}

bool IrcEngine::sendPart(const QString& chan, const QString& reason)
{
    return sendLine("PART "+chan+" :"+reason+"\r\n");
}

bool IrcEngine::sendQuote(const QString& text)
{
    return sendLine(text+"\r\n");
}

bool IrcEngine::sendStats(const QString& type)
{
    if(type.isEmpty())
        return sendLine("STATS\r\n");
    else
        return sendLine("STATS "+type+"\r\n");
}

bool IrcEngine::sendTopic(const QString& chan, const QString& topic)
{
    if (topic.isEmpty())
        return sendLine("TOPIC "+chan+"\r\n");
    else
        return sendLine("TOPIC "+chan+" :"+topic+"\r\n");
}

bool IrcEngine::sendTopic(const QString& chan)
{
    return sendLine("TOPIC "+chan+"\r\n");
}

bool IrcEngine::sendWallops(const QString& msg)
{
    return sendLine("WALLOPS :"+msg+"\r\n");
}

bool IrcEngine::sendVersion(const QString& to)
{
    return sendLine("NOTICE "+to+QString(" :\001VERSION dxirc (Qt toolkit) %1 (C) 2008~ by David Vachulka\001\r\n").arg(VERSION));
}

bool IrcEngine::sendWho(const QString& mask)
{
    return sendLine("WHO "+mask+"\r\n");
}

bool IrcEngine::sendWhoami()
{
    return sendLine("WHOIS "+m_nickName+"\r\n");
}

bool IrcEngine::sendWhois(const QString& params)
{
    return sendLine("WHOIS "+params+"\r\n");
}

bool IrcEngine::sendWhowas(const QString& params)
{
    return sendLine("WHOWAS "+params+"\r\n");
}

bool IrcEngine::sendLine(const QString& line)
{
    dxutils::debugLine(QString(">> %1").arg(line.section('\r',0,0)));
    int size=0;
    if(m_connected)
    {
        size = m_socket->write(line);
        if(size != -1) return true;
        else return false;
    }
    else return false;
}

void IrcEngine::sendCommands()
{
    if(!m_startCommands.contains('\n')) m_startCommands.append('\n');
    if(m_startCommands.right(1) != "\n") m_startCommands.append('\n');
    QStringList list = m_startCommands.split('\n', QString::SkipEmptyParts);
    for(int i=0; i < list.size(); i++)
    {
        sendCommand(list[i]);
    }
}

bool IrcEngine::sendCommand(const QString& commandtext)
{
    if(m_connected)
    {
        QString command = commandtext.section('/',1).section(' ',0,0).toLower();
        if(command == "admin")
        {
            return sendAdmin(commandtext.section(' ',1));
        }
        if(command == "away")
        {
            return sendAway(commandtext.section(' ',1));
        }
        if(command == "banlist")
        {
            QString channel = commandtext.section(' ',1);
            return sendBanlist(channel);
        }
        if(command == "ctcp")
        {
            QString to = commandtext.section(' ',1).section(' ',0,0);
            QString msg = commandtext.section(' ', 2);
            return sendCtcp(to, msg);
        }
        if(command == "deop")
        {
            QString params = commandtext.section(' ',1);
            QString channel = params.section(' ',0,0);
            QString nicks = params.section(' ',1);
            QString modeparams = dxutils::createModes('-', 'o', nicks);
            return sendMode(channel+" "+modeparams);
        }
        if(command == "devoice")
        {
            QString params = commandtext.section(' ',1);
            QString channel = params.section(' ',0,0);
            QString nicks = params.section(' ',1);
            QString modeparams = dxutils::createModes('-', 'v', nicks);
            return sendMode(channel+" "+modeparams);
        }
        if(command == "invite")
        {
            QString params = commandtext.section(' ',1);
            QString nick = params.section(' ',0,0);
            QString channel = params.section(' ',1);
            return sendInvite(nick, channel);
        }
        if(command == "join")
        {
            QString channel = commandtext.section(' ',1);
            return sendJoin(channel);
        }
        if(command == "kick")
        {
            QString params = commandtext.section(' ',1);
            QString channel = params.section(' ',0,0);
            QString nick = params.section(' ',1);
            QString reason = params.section(' ', 2);
            return sendKick(channel, nick, reason);
        }
        if(command == "kill")
        {
            QString params = commandtext.section(' ',1);
            QString nick = params.section(' ',0,0);
            QString reason = params.section(' ',1);
            return sendKill(nick, reason);
        }
        if(command == "list")
        {
            return sendList(commandtext.section(' ',1));
        }
        if(command == "me")
        {
            QString params = commandtext.section(' ',1);
            QString to = params.section(' ',0,0);
            QString message = params.section(' ',1);
            return sendMe(to, message);
        }
        if(command == "msg")
        {
            QString params = commandtext.section(' ',1);
            QString to = params.section(' ',0,0);
            QString message = params.section(' ',1);
            return sendMsg(to, message);
        }
        if(command == "names")
        {
            QString params = commandtext.section(' ',1);
            return sendNames(params, false);
        }
        if(command == "nick")
        {
            QString nick = commandtext.section(' ',1);
            return sendNick(nick);
        }
        if(command == "notice")
        {
            QString params = commandtext.section(' ',1);
            QString to = params.section(' ',0,0);
            QString message = params.section(' ',1);
            return sendNotice(to, message);
        }
        if(command == "op")
        {
            QString params = commandtext.section(' ',1);
            QString channel = params.section(' ',0,0);
            QString nicks = params.section(' ',1);
            QString modeparams = dxutils::createModes('+', 'o', nicks);
            return sendMode(channel+" "+modeparams);
        }
        if(command == "oper")
        {
            QString params = commandtext.section(' ',1);
            QString login = params.section(' ',0,0);
            QString password = params.section(' ',1);
            return sendOper(login, password);
        }
        if(command == "quote")
        {
            return sendQuote(commandtext.section(' ',1));
        }
        if(command == "topic")
        {
            QString params = commandtext.section(' ',1);
            QString channel = params.section(' ',0,0);
            QString topic = params.section(' ',1);
            return sendTopic(channel, topic);
        }
        if(command == "voice")
        {
            QString params = commandtext.section(' ',1);
            QString channel = params.section(' ',0,0);
            QString nicks = params.section(' ',1);
            QString modeparams = dxutils::createModes('+', 'v', nicks);
            return sendMode(channel+" "+modeparams);
        }
        if(command == "wallops")
        {
            QString params = commandtext.section(' ',1);
            return sendWallops(params);
        }
        if(command == "who")
        {
            QString params = commandtext.section(' ',1);
            return sendWho(params);
        }
        if(command == "whois")
        {
            QString params = commandtext.section(' ',1);
            return sendWhois(params);
        }
        if(command == "whowas")
        {
            QString params = commandtext.section(' ',1);
            return sendWhowas(params);
        }
        emit ircError(_("Bad command on connection:%1").arg(commandtext));
        return false;
    }
    else
        return false;
}

QString IrcEngine::getLocalIPBinary()
{
    return m_socket->stringIPToBinary(getLocalIP());
}

QString IrcEngine::getHostname(const QString& address)
{
    return m_socket->getHostname(address);
}

void IrcEngine::clearChannelsCommands(bool disableReconnect)
{
    if(preferences.m_reconnect && hasReconnectAttempt() && !disableReconnect)
    {
        makeStartChannels();
    }
    else
    {
        m_startChannels.clear();
        m_startCommands.clear();
    }
}

void IrcEngine::makeStartChannels()
{
    m_startChannels = m_channels.join(",");
}

void IrcEngine::addNick(const QString &nick, const QString &user, const QString &real, const QString &host, const bool &away)
{
    bool nickExist = false;
    for(int i=0; i < m_nicks.size(); i++)
    {
        if(m_nicks[i].nick == nick)
        {
            m_nicks[i].user = user;
            m_nicks[i].real = real;
            m_nicks[i].host = host;
            m_nicks[i].away = away;
            nickExist = true;
            emit ircNickInfo(m_nicks[i]);
            break;
        }
    }
    if(!nickExist)
    {
        NickInfo nickStruct;
        nickStruct.nick = nick;
        nickStruct.user = user;
        nickStruct.real = real;
        nickStruct.host = host;
        nickStruct.away = away;
        m_nicks.append(nickStruct);
        emit ircNickInfo(nickStruct);
    }
}

void IrcEngine::addNick(const QString &nick, const QString &user, const QString &real, const QString &host)
{
    bool nickExist = false;
    for(int i=0; i < m_nicks.size(); i++)
    {
        if(m_nicks[i].nick == nick)
        {
            m_nicks[i].user = user;
            m_nicks[i].real = real;
            m_nicks[i].host = host;
            nickExist = true;
            emit ircNickInfo(m_nicks[i]);
            break;
        }
    }
    if(!nickExist)
    {
        NickInfo nickStruct;
        nickStruct.nick = nick;
        nickStruct.user = user;
        nickStruct.real = real;
        nickStruct.host = host;
        nickStruct.away = false;
        m_nicks.append(nickStruct);
        emit ircNickInfo(nickStruct);
    }
}

void IrcEngine::removeNick(const QString &nick)
{
    for(int i=0; i < m_nicks.size(); i++)
    {
        if(m_nicks[i].nick == nick)
        {
            m_nicks.removeAt(i);
            break;
        }
    }
}

NickInfo IrcEngine::getNickInfo(const QString &nick) const
{
    for(int i=0; i < m_nicks.size(); i++)
    {
        if(m_nicks[i].nick == nick)
        {
            return m_nicks[i];
        }
    }
    NickInfo nickInfo;
    nickInfo.away = false;
    return nickInfo;
}

QStringList IrcEngine::getBannedNick(const QString &banmask) const
{
    QStringList nicks;
    QString banNick = banmask.section('!',0,0);
    QString banUser = banmask.section('!',1).section('@',0,0);
    QString banHost = banmask.section('@',1);
    for(int i=0; i<m_nicks.size(); i++)
    {
        if(QRegExp(QString("\\<"+banNick+"\\>").replace("*","\\w*")).exactMatch(m_nicks[i].nick) && QRegExp(QString(banUser+"\\>").replace("*","\\w*")).exactMatch(m_nicks[i].user) && QRegExp(QString("\\<"+banHost+"\\>").replace("*","\\w*")).exactMatch(m_nicks[i].host))
            nicks.append(m_nicks[i].nick);
    }
    return nicks;
}

QStringList IrcEngine::getChannelsList() const
{
    return m_listChannels;
}

void IrcEngine::clearChannelsList()
{
    m_listChannels.clear();
}

bool IrcEngine::isAway(const QString& nick)
{
    for(int i=0; i < m_nicks.size(); i++)
    {
        if(m_nicks[i].nick == nick)
        {
            return m_nicks[i].away;
        }
    }
    return false;
}

bool IrcEngine::isUserIgnored(QString nick, QString user, QString host, QString on)
{
    bool bnick = false;
    bool buser = false;
    bool bhost = false;
    bool bchannel = false;
    bool bnetwork = false;
    for(int i=0; i<preferences.m_ignoreUsersList.size(); i++)
    {
        QString inick, iuser, ihost;
        inick = preferences.m_ignoreUsersList[i].nick.section('!',0,0).toLower();
        iuser = preferences.m_ignoreUsersList[i].nick.section('!',1).section('@',0,0).toLower();
        ihost = preferences.m_ignoreUsersList[i].nick.section('@',1).toLower();
        if(QRegExp(QString("\\<"+inick+"\\>").replace("*","\\w*")).exactMatch(nick.toLower())) bnick = true;
        if(QRegExp(QString("\\<"+iuser+"\\>").replace("*","\\w*")).exactMatch(user.toLower()) || iuser.isEmpty()) buser = true;
        if(QRegExp(QString("\\<"+ihost+"\\>").replace("*","\\w*")).exactMatch(host.toLower()) || ihost.isEmpty()) bhost = true;
        if(bnick && buser && bhost)
        {
            if(preferences.m_ignoreUsersList[i].channel == "all") bchannel = true;
            if(preferences.m_ignoreUsersList[i].channel.contains(','))
            {
                for(int j=1; j<preferences.m_ignoreUsersList[i].channel.count(',')+2; j++)
                {
                    if(QRegExp(QString(dxutils::getParam(preferences.m_ignoreUsersList[i].channel, j, false, ',')+"\\>").toLower().replace("*","\\w*")).exactMatch(on.toLower()))
                    {
                        bchannel = true;
                        break;
                    }
                }
            }
            else
            {
                if(QRegExp(QString(preferences.m_ignoreUsersList[i].channel+"\\>").toLower().replace("*","\\w*")).exactMatch(on.toLower())) bchannel = true;
            }
            if(preferences.m_ignoreUsersList[i].network == "all") bnetwork = true;
            if(QRegExp(QString("\\<"+preferences.m_ignoreUsersList[i].network+"\\>").toLower().replace("*","\\w*")).exactMatch(getNetworkName().toLower())) bnetwork = true;
            break;
        }
    }
    return bnick && buser && bhost && bchannel && bnetwork;
}

bool IrcEngine::isUserIgnored(QString nick, QString on)
{
    QString user, host;
    for(int i = 0; i < m_nicks.size(); i++)
    {
        if(nick == m_nicks[i].nick)
        {
            user = m_nicks[i].user;
            host = m_nicks[i].host;
            break;
        }
    }
    return isUserIgnored(nick, user, host, on);
}

// reset running reconnect timeout
void IrcEngine::resetReconnect()
{
    if(m_timerReconnect->isActive())
    {
        m_timerReconnect->stop();
        m_timerReconnect->start(preferences.m_delayAttempt*1000);
        return;
    }
    if(preferences.m_reconnect && hasReconnectAttempt())
    {
        m_timerReconnect->start(preferences.m_delayAttempt*1000);
    }
}

//send ping for question connection alive
void IrcEngine::sendPingTimeout()
{
    if(m_connected)
    {
        m_ping = QDateTime::currentDateTime().currentMSecsSinceEpoch();
        sendLine(QString("PING %1\r\n").arg(m_ping));
        m_timerPing->start(60000);
    }
}

//can we reconnect attempt
bool IrcEngine::hasReconnectAttempt()
{
    if(!preferences.m_numberAttempt) return true;
    if(m_attempts < preferences.m_numberAttempt) return true;
    return false;
}

