/*
 *      dcctabitem.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include <string>
#include "dcctabitem.h"
#include "dxutils.h"
#include "config.h"
#include "i18n.h"
#include "preferences.h"

static bool caseInsensitiveLessThan(const QString &s1, const QString &s2)
{
    return s1.toLower() < s2.toLower();
}

DccTabItem::DccTabItem(TabWidget *tab, const QString &mynick, const QString &nick,
        const QString &address, int portD, int portH, bool listen,
        int id, IrcEngine *engine, QWidget *parent)
    : dxTabItem(tab, id, parent), m_listen(listen),
        m_logstream(NULL), m_mynick(mynick), m_nick(nick),
        m_address(address), m_portD(portD), m_portH(portH), m_engine(engine)
{
    m_currentPosition = 0;
    m_historyMax = 25;
    m_pipe = NULL;
    m_sendPipe = false;
    m_scriptHasAll = false;
    m_scriptHasOwnMsg = false;
    m_receiveRest = "";
    m_dccIP = dxUtils.getStringIniEntry("SETTINGS", "dccIP");
    m_dccTimeout = dxUtils.getIntIniEntry("SETTINGS", "dccTimeout", 66);

    m_pipeTimer = new QTimer(this);
    m_closeTimer = new QTimer(this);
    m_text = new QTextBrowser;
    m_text->setOpenExternalLinks(true);
    m_text->document()->setDefaultStyleSheet(QString("a { color:%1 }").arg(preferences.m_colors.link.name()));
    m_text->setStyleSheet(QString("QTextEdit { color:%1; background:%2; }").arg(preferences.m_colors.text.name()).arg(preferences.m_colors.back.name()));
    m_text->setFont(preferences.m_ircFont);
    m_commandline = new LineSpell;
    m_commandline->installEventFilter(this);
    m_commandline->setStyleSheet(QString("QTextEdit { color:%1; background:%2; }").arg(preferences.m_colors.text.name()).arg(preferences.m_colors.back.name()));
    if(preferences.m_sameCmd) m_commandline->setFont(preferences.m_ircFont);
    m_spellLangs = new QComboBox;
    m_spellLangs->setToolTip(_("Spellchecking language list"));
    m_spellLangs->hide();
    if(preferences.m_useSpell && dxUtils.getAvailableSpellLangsNum())
    {
        QStringList langs = dxUtils.getAvailableSpellLangs();
        QString lang = dxUtils.getChannelLang(m_nick);
        for(int i=0; i<langs.size(); ++i)
        {
            m_spellLangs->addItem(langs[i]);
            if(langs[i]==lang) m_spellLangs->setCurrentIndex(i);;
        }
        if(preferences.m_showSpellCombo) m_spellLangs->show();
        m_commandline->setUseSpell(true);
        m_commandline->setLanguage(lang);
        m_commandline->setToolTip(_("Current spellchecking language: %1").arg(lang));
    }
    QHBoxLayout *command = new QHBoxLayout;
    command->setSpacing(0);
    command->addWidget(m_commandline);
    command->addWidget(m_spellLangs);
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    mainLayout->addWidget(m_text);
    mainLayout->addLayout(command);
    setLayout(mainLayout);

    m_socket = new dxSocket();

    connect(m_commandline, SIGNAL(returnPressed()), this, SLOT(slotProcessCommandline()));
    connect(m_pipeTimer, SIGNAL(timeout()), this, SLOT(slotPipeTimeout()));
    connect(m_closeTimer, SIGNAL(timeout()), this, SLOT(slotCloseTimeout()));
    if(m_engine) connect(m_engine, SIGNAL(ircNick(QString,QString)), this, SLOT(slotIrcNick(QString,QString)));
    connect(m_socket, SIGNAL(socketConnected()), this, SLOT(slotConnected()));
    connect(m_socket, SIGNAL(socketDisconnected()), this, SLOT(slotDisconnected()));
    connect(m_socket, SIGNAL(socketCanRead()), this, SLOT(slotCanRead()));
    connect(m_socket, SIGNAL(socketError(QString)), this, SLOT(slotError(QString)));
    connect(m_socket, SIGNAL(socketStartAccept(int)), this, SLOT(slotStartAccept(int)));
    connect(m_socket, SIGNAL(socketSomeConnected()), this, SLOT(slotSomeConnected()));

    start();
}

DccTabItem::~DccTabItem()
{
    this->stopLogging();
    if(m_pipe) m_pipe->close();
    m_pipeStrings.clear();
    m_pipeTimer->stop();
    m_socket->disconnectFrom();
    delete m_socket;
}

void DccTabItem::start()
{
    m_commandline->setFocus();
    m_socket->setIsSSL(false);
    if(m_listen)
    {
        writeEvent(_("Listening on %1").arg(m_address), CONNECT);
        if(m_socket->isIPv4(m_address))
        {
            if(!m_socket->isRoutableIP(m_socket->stringIPToBinary(m_address).toUInt()))
            {
                writeEvent(_("%1 isn't routable.").arg(m_address), DCCERROR);
            }
            if(!m_engine->getMyUserHost().isEmpty() || !m_dccIP.isEmpty())
            {
                if(!m_dccIP.isEmpty())
                {
                    m_address = m_dccIP;
                    writeEvent(_("Trying IP from settings %1").arg(m_address), CONNECT);
                }
                else
                {
                    m_address = m_engine->getMyUserHost();
                    writeEvent(_("Trying IP from server %1").arg(m_address), CONNECT);
                }
                if(m_socket->isIPv4(m_address) && !m_socket->isRoutableIP(m_socket->stringIPToBinary(m_address).toUInt()))
                {
                    writeEvent(_("%1 isn't routable too, but will be used").arg(m_address), DCCERROR);
                }
            }
            else
            {
                writeEvent(_("But better doesn't exist."), DCCERROR);
            }
        }
    }
    m_socket->setHost(m_address, m_portD, m_portH);
    if(m_listen) m_socket->listenOn();
    else m_socket->connectTo();
    loadLogLines();
}

void DccTabItem::clearChat()
{
    m_text->clear();
}

//usefull for set tab current
void DccTabItem::makeLastRowVisible()
{
    QTextCursor c =  m_text->textCursor();
    c.movePosition(QTextCursor::End);
    m_text->setTextCursor(c);
}

QString DccTabItem::getSpellLang()
{
#ifdef HAVE_ENCHANT
    if(m_spellLangs->count()) return m_spellLangs->currentText();
#endif
    return "";
}

void DccTabItem::preferencesUpdated()
{
    m_text->document()->setDefaultStyleSheet(QString("a { color:%1 }").arg(preferences.m_colors.link.name()));
    m_text->setStyleSheet(QString("QTextEdit { color:%1; background:%2; }").arg(preferences.m_colors.text.name()).arg(preferences.m_colors.back.name()));
    m_commandline->setStyleSheet(QString("QLineEdit { color:%1; background:%2; }").arg(preferences.m_colors.text.name()).arg(preferences.m_colors.back.name()));
    m_text->setFont(preferences.m_ircFont);
    if(preferences.m_sameCmd) m_commandline->setFont(preferences.m_ircFont);
    else m_commandline->setFont(QFont());
    stopLogging();
    if(preferences.m_useSpell && dxUtils.getAvailableSpellLangsNum())
    {
        QStringList langs = dxUtils.getAvailableSpellLangs();
        QString lang = dxUtils.getChannelLang(m_nick);
        for(int i=0; i<langs.size(); ++i)
        {
            m_spellLangs->addItem(langs[i]);
            if(langs[i]==lang) m_spellLangs->setCurrentIndex(i);
        }
        if(preferences.m_showSpellCombo) m_spellLangs->show();
        m_commandline->setUseSpell(true);
        m_commandline->setLanguage(lang);
        m_commandline->setToolTip(_("Current spellchecking language: %1").arg(lang));
    }
    else
    {
        m_commandline->setUseSpell(false);
        m_commandline->setToolTip("");
        m_spellLangs->hide();
    }
}

void DccTabItem::showSpellComboUpdated()
{
    if(preferences.m_showSpellCombo) m_spellLangs->show();
    else m_spellLangs->hide();
}

//if highlight==true, highlight tab
void DccTabItem::appendText(QString msg, bool bhighlight, bool logLine)
{
    appendIrcText(msg, 0, false, logLine);
    if(bhighlight)
    {
        if(highlightNeeded(msg))
        {
            emit highlight(DCC);
        }
        else emit newmsg(DCC);
    }
}

void DccTabItem::appendIrcText(QString msg, QString time, bool disableStrip, bool logLine)
{
    if(time.isEmpty()) time = QDateTime::currentDateTime().toString("[hh:mm:ss] ");
    if(preferences.m_stripColors && !disableStrip) msg = dxutils::stripColors(msg, false);
    m_text->append("<div>"+time+dxutils::parseText(msg, 0)+"</div>");
    if(logLine) this->logLine(msg, time);
}

void DccTabItem::appendIrcNickText(QString nick, QString msg, QString time, bool logLine)
{
    if(time.isEmpty()) time = QDateTime::currentDateTime().toString("[hh:mm:ss] ");
    if(preferences.m_stripColors) msg = dxutils::stripColors(msg, false);
    QString nickStyled = "";
    if(preferences.m_coloredNick) nickStyled = QString("<font color=\"%1\">%2: </font>").arg(dxutils::getNickColor(nick)).arg(nick);
    else nickStyled = QString("<b>%1: </b>").arg(nick);
    m_text->append("<div>"+time+nickStyled+dxutils::parseText(msg, 0)+"</div>");
    if(logLine) this->logLine("<"+nick+"> "+msg, time);
}

void DccTabItem::appendMyMsg(QString msg, QString time)
{
    if(time.isEmpty()) time = QDateTime::currentDateTime().toString("[hh:mm:ss] ");
    if(preferences.m_stripColors) msg = dxutils::stripColors(msg, false);
    QString nick = "";
    if(preferences.m_coloredNick) nick = QString("<font color=\"%1\">%2: </font>").arg(dxutils::getColorForStyle(10).name()).arg(getNickName());
    else nick = QString("<b>%1: </b>").arg(getNickName());
    m_text->append("<div>"+time+nick+dxutils::parseText(msg, 10)+"</div>");
    this->logLine("\01410\014<"+getNickName()+"> "+msg, time);
}

/* if highlight==true, highlight tab
 * disableStrip is for dxirc.Print
*/
void DccTabItem::appendStyledText(QString text, int style, bool bhighlight, bool disableStrip, bool logLine)
{
    if(style) appendIrcStyledText(text, style, "", disableStrip, logLine);
    else appendIrcText(text, "", disableStrip, logLine);
    if(bhighlight)
    {
        if(highlightNeeded(text)) emit highlight(DCC);
        else emit newmsg(DCC);
    }
}

void DccTabItem::appendIrcStyledText(QString styled, int stylenum, QString time, bool disableStrip, bool logLine)
{
    if(time.isEmpty()) time = QDateTime::currentDateTime().toString("[hh:mm:ss] ");
    if(preferences.m_stripColors && !disableStrip) styled = dxutils::stripColors(styled, false);
    m_text->append("<div>"+time+dxutils::parseText(styled, stylenum)+"</div>");
    if(logLine)
    {
        if(stylenum==8) //highlight msg
        {
            this->logLine("\0148\014<"+styled.section(':',0,0)+">"+styled.section(':',1), time);
        }
        else
        {
            this->logLine(QString("\014%1\014%2").arg(stylenum).arg(styled), time);
        }
    }
}
void DccTabItem::startLogging()
{
    if(m_logstream && QFile::exists(preferences.m_logPath+"/"+getServerName()+"/"+m_nick+"/"+QDateTime::currentDateTime().toString("yyyy-MM-dd")))
        return;
    if(preferences.m_logging)
    {
        QDir dir;
        if(!dir.exists(preferences.m_logPath+"/"+getServerName()))
            dir.mkpath(preferences.m_logPath+"/"+getServerName());
        if(!dir.exists(preferences.m_logPath+"/"+getServerName()+"/"+m_nick))
            dir.mkpath(preferences.m_logPath+"/"+getServerName()+"/"+m_nick);
        m_logstream = new std::ofstream(dxutils::toStdString(preferences.m_logPath+QDir::separator()+getServerName()+QDir::separator()+m_nick+QDir::separator()+QDateTime::currentDateTime().toString("yyyy-MM-dd")).c_str(), std::ios::out|std::ios::app);
    }
}

void DccTabItem::stopLogging()
{
    if(m_logstream)
    {
        m_logstream->close();
        delete m_logstream;
        m_logstream = NULL;
    }
}

void DccTabItem::logLine(const QString &line, const QString &time)
{
    if(preferences.m_logging)
    {
        this->startLogging();
        *m_logstream << dxutils::toStdString(time) << dxutils::toStdString(line) << std::endl;
    }
}

void DccTabItem::slotProcessCommandline()
{
    QString commandtext = m_commandline->text();
    if(commandtext.isEmpty())
        return;
    m_commandsHistory.append(commandtext);
    if (m_commandsHistory.size() > m_historyMax)
        m_commandsHistory.removeFirst();
    m_currentPosition = m_commandsHistory.size()-1;
    m_commandline->clear();
    if(QString::compare(commandtext.left(4),"/say", Qt::CaseInsensitive) != 0)
        commandtext.replace("^B", "\002").replace("^C", "\003").replace("^O", "\017").replace("^V", "\026").replace("^_", "\037");
    for(int i=0; i<=commandtext.count('\n'); i++)
    {
        QString text = commandtext.section('\n', i).section('\r',0,0);
        if(QString::compare(text.section('/',1).section(' ',0,0), "quit", Qt::CaseInsensitive) == 0 ||
                QString::compare(text.section('/',1).section(' ',0,0), "lua", Qt::CaseInsensitive) == 0)
        {
            processLine(text);
            return;
        }
#ifdef HAVE_LUA
        if(text[0] != '/') emit ownMsg(text);
        emit command(text);
        if(text[0] == '/' && !m_scriptHasAll) processLine(text);
        else if(!m_scriptHasOwnMsg && !m_scriptHasAll) processLine(text);
#else
        processLine(text);
#endif
    }
}

bool DccTabItem::processLine(const QString& commandtext)
{
    return processCommand(commandtext);
}

bool DccTabItem::processCommand(const QString& commandtext)
{
    QString command = (commandtext[0] == '/' ? commandtext.section('/',1).section(' ',0,0).toLower() : "");
    if(dxUtils.isScriptCommand(command))
    {
        LuaRequest lua;
        lua.type = LUA_COMMAND;
        lua.text = commandtext.section('/',1);
        emit luaRequest(lua);
        return true;
    }
    if(command == "commands")
    {
        appendIrcStyledText(dxUtils.availableDccCommands(), 3, "", false, false);
        return true;
    }
    if(command == "exec")
    {
        QString params = commandtext.section(' ',1);
        if(params.isEmpty())
        {
            appendIrcStyledText(_("/exec [-o|-c] <command>, executes command, -o sends output to channel/query, -c closes running command."), 4, "", false, false);
            return false;
        }
        else
        {
            if(!m_pipe)
            {
                m_pipe = new QProcess(this);
                connect(m_pipe, SIGNAL(readyReadStandardOutput()), this, SLOT(slotOnPipe()));
            }
            m_pipeStrings.clear();
            if(params.section(' ',0,0).contains("-o"))
            {
                m_sendPipe = true;
                m_pipe->start(params.section(' ',1));
            }
            else if(params.section(' ',0,0).contains("-c"))
            {
                m_sendPipe = false;
                m_pipeStrings.clear();
                m_pipe->close();
            }
            else
            {
                m_sendPipe = false;
                m_pipe->start(params);
            }
            return true;
        }
    }
    if(command == "help")
    {
        return showHelp(commandtext.section(' ',1).toLower().trimmed());
    }
    if(command == "log")
    {
        QString logcommand = commandtext.section(' ',1).section(' ',0,0);
        QString logfile = commandtext.section(' ',1).section(' ',1);
        if(QString::compare(logcommand, "list", Qt::CaseInsensitive)==0)
        {
            listLogFiles();
            return true;
        }
        else if(QString::compare(logcommand, "show", Qt::CaseInsensitive)==0)
        {
            if(logfile.isEmpty())
                logfile = QDateTime::currentDateTime().toString("yyyy-MM-dd");
            QString logpath = preferences.m_logPath+"/"+getServerName()+"/"+m_nick+"/"+logfile;
            if(!QFile::exists(logpath))
            {
                appendIrcStyledText(_("Log file %1 doesn't exist").arg(logfile), 4, "", false, false);
                return false;
            }
            emit newlog(logpath);
            return true;
        }
        else
        {
            appendIrcStyledText(_("'%1' isn't <list|show>").arg(logcommand), 4, "", false, false);
            return false;
        }
    }
    if(command == "lua")
    {
#ifdef HAVE_LUA
        QString luacommand = commandtext.section(' ',1).section(' ',0,0);
        QString luatext = commandtext.section(' ',1).section(' ',1);
        LuaRequest lua;
        if(luacommand.isEmpty())
        {
            appendIrcStyledText(_("/lua <help|load|unload|list> [scriptpath|scriptname]"), 4, "", false, false);
            return false;
        }
        if(QString::compare(luacommand, "help", Qt::CaseInsensitive)==0)
        {
            appendIrcStyledText(_("For help about Lua scripting visit: %1").arg(LUA_HELP_PATH), 3, "", false, false);
            return true;
        }
        else if(QString::compare(luacommand, "load", Qt::CaseInsensitive)==0) lua.type = LUA_LOAD;
        else if(QString::compare(luacommand, "unload", Qt::CaseInsensitive)==0) lua.type = LUA_UNLOAD;
        else if(QString::compare(luacommand, "list", Qt::CaseInsensitive)==0) lua.type = LUA_LIST;
        else
        {
            appendIrcStyledText(_("'%1' isn't <help|load|unload|list>").arg(luacommand), 4, "", false, false);
            return false;
        }
        lua.text = luatext;
        emit luaRequest(lua);
        return true;
#else
        appendIrcStyledText(_("dxirc is compiled without support for Lua scripting"), 4, "", false, false);
        return false;
#endif
    }
    if(command == "me")
    {
        if(m_socket->isConnected())
        {
            appendIrcStyledText(m_mynick+" "+commandtext.section(' ',1), 2, "");
            return sendLine("\001ACTION "+commandtext.section(' ',1)+"\001\n");
        }
        else
        {
            appendIrcStyledText(_("You aren't connected"), 4, "", false, false);
            return true;
        }
    }
    if(command == "say")
    {
        if(m_socket->isConnected())
        {
            appendMyMsg(commandtext.section(' ',1), "");
            return sendLine(commandtext.section(' ',1)+"\n");
        }
        else
        {
            appendIrcStyledText(_("You aren't connected"), 4, "", false, false);
            return true;
        }
    }
    if(command == "tetris")
    {
        void emit newtetris();
        return true;
    }
    if(commandtext[0] != '/')
    {
        if(m_socket->isConnected())
        {
            appendMyMsg(commandtext, "");
            return sendLine(commandtext+"\n");
        }
        else
        {
            appendIrcStyledText(_("You aren't connected"), 4, "", false, false);
            return true;
        }
    }
    return false;
}

bool DccTabItem::showHelp(QString command)
{
    if(dxUtils.isScriptCommand(command))
    {
        appendIrcStyledText(dxUtils.getHelpText(command), 3, "", false, false);
        return true;
    }
    if(command == "commands")
    {
        appendIrcStyledText(_("COMMANDS, shows available commands"), 3, "", false, false);
        return true;
    }
    if(command == "exec")
    {
        appendIrcStyledText(_("EXEC [-o|-c] <command>, executes command, -o sends output to channel/query, -c closes running command."), 3, "", false, false);
        return true;
    }
    if(command == "help")
    {
        appendIrcStyledText(_("HELP <command>, shows help for command."), 3, "", false, false);
        return true;
    }
    if(command == "log")
    {
        appendIrcStyledText(_("LOG list, shows list of log files for current tab"), 3, "", false, false);
        appendIrcStyledText(_("LOG show [date], shows last log file for current tab, if date is empty"), 3, "", false, false);
        appendIrcStyledText(_("else file for given date"), 3, "", false, false);
        appendIrcStyledText(_("date example: 2011-01-31"), 3, "", false, false);
        return true;
    }
#ifdef HAVE_LUA
    if(command == "lua")
    {
        appendIrcStyledText(_("LUA help, shows help for lua scripting."), 3, "", false, false);
        appendIrcStyledText(_("LUA load <path>, loads script."), 3, "", false, false);
        appendIrcStyledText(_("Example: /lua load /home/dvx/test.lua"), 3, "", false, false);
        appendIrcStyledText(_("LUA unload <name>, unloads script."), 3, "", false, false);
        appendIrcStyledText(_("Example: /lua unload test"), 3, "", false, false);
        appendIrcStyledText(_("LUA list, shows list of loaded scripts"), 3, "", false, false);
        return true;
    }
#else
    if(command == "lua")
    {
        appendIrcStyledText(_("dxirc is compiled without support for Lua scripting"), 4, "", false, false);
        return true;
    }
#endif
    if(command == "me")
    {
        appendIrcStyledText(_("ME <to> <message>, sends the action."), 3, "", false, false);
        return true;
    }
    if(command == "say")
    {
        appendIrcStyledText(_("SAY [text], sends text to current tab."), 3, "", false, false);
        return true;
    }
    if(command == "tetris")
    {
        appendIrcStyledText(_("TETRIS, start small easteregg."), 3, "", false, false);
        appendIrcStyledText(_("Keys for playing:"), 3, "", false, false);
        appendIrcStyledText(_("n .. new game"), 3, "", false, false);
        appendIrcStyledText(_("p .. pause game"), 3, "", false, false);
        appendIrcStyledText(_("i .. rotate piece"), 3, "", false, false);
        appendIrcStyledText(_("l .. move piece right"), 3, "", false, false);
        appendIrcStyledText(_("k .. drop piece"), 3, "", false, false);
        appendIrcStyledText(_("j .. move piece left"), 3, "", false, false);
        return true;
    }
    if(!dxUtils.getAlias(command[0] == '/' ? command:"/"+command).isEmpty())
    {
        appendIrcStyledText(QString("%1: %2").arg(command.toUpper()).arg(dxUtils.getAlias(command[0] == '/' ? command:"/"+command)), 3, "", false, false);
        return true;
    }
    if(command.isEmpty()) appendIrcStyledText(_("Command is empty, type /commands for available commands"), 4, "", false, false);
    else appendIrcStyledText(_("Unknown command '%1', type /commands for available commands").arg(command), 4, "", false, false);
    return false;
}


void DccTabItem::slotOnPipe()
{
    if(!m_socket->isConnected())
    {
        m_sendPipe = false;
        m_pipeStrings.clear();
        m_pipe->close();
        return;
    }
    QByteArray byteArray = m_pipe->readAllStandardOutput();
    bool isUtf8 = dxutils::isUtf8(byteArray);
    QStringList lines = QString(isUtf8?QString::fromUtf8(byteArray):QString::fromLocal8Bit(byteArray)).split('\n', QString::SkipEmptyParts);
    foreach(QString line, lines)
    {
        if(m_sendPipe)
        {
            if(!m_pipeTimer->isActive()) m_pipeTimer->start(3000);
            m_pipeStrings.append(line);
        }
        else appendIrcText(line, "", true, false);
    }
}

void DccTabItem::slotIrcNick(QString nick, QString newnick)
{
    if(m_mynick==nick)
    {
        m_mynick = newnick;
        appendIrcStyledText(_("You're now known as %1").arg(m_mynick), 1, "", false, false);
    }
    if(m_nick==nick)
    {
        m_nick = newnick;
        appendIrcStyledText(_("%1 is now known as %2").arg(nick).arg(newnick), 1, "", false, false);
        emit newname(m_nick);
        stopLogging();
    }
}

void DccTabItem::slotPipeTimeout()
{
    if(m_pipeStrings.size() > 3)
    {
        if(m_pipeStrings[0].isEmpty()) m_pipeStrings[0] = " ";
#ifdef HAVE_LUA
        emit ownMsg(m_pipeStrings[0]);
        if(!m_scriptHasOwnMsg)
        {
            appendMyMsg(m_pipeStrings[0], "");
            sendLine(m_pipeStrings[0]+"\n");
        }
#else
        appendMyMsg(m_pipeStrings[0], "");
        sendLine(m_pipeStrings[0]+"\n");
#endif
        m_pipeStrings.pop_front();
        m_pipeTimer->start(3000);
    }
    else
    {
        while(m_pipeStrings.size())
        {
            if(m_pipeStrings[0].isEmpty()) m_pipeStrings[0] = " ";
#ifdef HAVE_LUA
            emit ownMsg(m_pipeStrings[0]);
            if(!m_scriptHasOwnMsg)
            {
                appendMyMsg(m_pipeStrings[0], "");
                sendLine(m_pipeStrings[0]+"\n");
            }
#else
            appendMyMsg(m_pipeStrings[0], "");
            sendLine(m_pipeStrings[0]+"\n");
#endif
            m_pipeStrings.pop_front();
        }
    }
}

//for offered connectin to me
void DccTabItem::slotCloseTimeout()
{
    if(!m_socket->isConnected())
    {
        m_closeTimer->stop();
        appendIrcStyledText(_("Connection closed. Client didn't connect in given timeout"), 4, "", true, false);
        m_socket->disconnectFrom();
    }
}

QStringList DccTabItem::cutText(QString text, int len)
{
    register int previous = 0;
    register int s = -1;
    register int lw = 0;
    QStringList texts;
    for(int i = 0; i<text.length(); i++)
    {
        if(text[i] == ' ') s = i;
        if(lw>len)
        {
            if(s!=-1)
            {
                texts.append(text.mid(previous, s));
                i = s+1;
                s = -1;
                previous = i;
            }
            else
            {
                texts.append(text.mid(previous, lw));
                previous = previous+lw;
            }
            lw = 0;
        }
        lw++;
    }
    if(previous<text.length()) texts.append(text.right(text.length()-previous));
    return texts;
}

void DccTabItem::setCommandFocus()
{
    m_commandline->setFocus();
}

//for "handle" checking, if script contains "all". Send from dxirc.
void DccTabItem::setHasAllCommand(bool hasAll)
{
    m_scriptHasAll = hasAll;
}

//for "handle" checking, if script contains "ownmsg". Send from dxirc.
void DccTabItem::setHasOwnMsg(bool hasOwn)
{
    m_scriptHasOwnMsg = hasOwn;
}

//check need of highlight in msg
bool DccTabItem::highlightNeeded(const QString &msg)
{
    int pos = msg.indexOf(getNickName());
    if(pos==-1) return false;
    bool before = true;
    bool after = false;
    if(pos) before = dxutils::isDelimiter(msg[pos-1].toLatin1());
    if(pos+getNickName().length() == msg.length()) after = true;
    if(pos+getNickName().length() < msg.length()) after = dxutils::isDelimiter(msg[pos+getNickName().length()].toLatin1());
    return before && after;
}

bool DccTabItem::sendLine(const QString& line)
{
    dxutils::debugLine(QString(">> %1").arg(line.section('\n',0,0)));
    int size=0;
    if(m_socket->isConnected())
    {
        size = m_socket->write(line.toLocal8Bit(), line.length());
        if(size != -1) return true;
        else return false;
    }
    else return false;
}

void DccTabItem::slotCanRead()
{
    char buffer[1024];
    QString data = m_receiveRest;
    int size = 0;
    while((size = m_socket->read(buffer, 1023)) > 0)
    {
        buffer[size] = '\0';
        if(dxutils::isUtf8(buffer)) data.append(QString::fromUtf8(buffer));
        else data.append(QString::fromLocal8Bit(buffer));
        while(data.contains('\n'))
        {
            QString msg = data.section('\n',0,0).section('\r',0,0);
            Event ev;
            if(msg.contains('\001'))
            {
                ev = DCCACTION;
                msg = msg.section('\001',1).section('\001',0,0).section(' ',1);
            }
            else
                ev = DCCMSG;
            dxutils::debugLine(QString("<< %1").arg(msg));
            writeEvent(msg, ev);
            data = data.section('\n',1);
        }
        m_receiveRest = data;
    }
}

void DccTabItem::slotConnected()
{
    writeEvent(_("Connected to %1").arg(m_address), CONNECT);
}

void DccTabItem::slotDisconnected()
{
    m_sendPipe = false;
    m_pipeStrings.clear();
    if(m_pipe) m_pipe->close();
    writeEvent(_("Server %1 was disconnected").arg(m_address), DISCONNECT);
}

void DccTabItem::slotError(QString error)
{
    m_sendPipe = false;
    m_pipeStrings.clear();
    if(m_pipe) m_pipe->close();
    writeEvent(error, DCCERROR);
}

void DccTabItem::slotStartAccept(int port)
{
    m_portD = port;
    if(m_engine && m_listen) m_engine->sendCtcp(m_nick, "DCC CHAT chat "+m_socket->stringIPToBinary(m_address)+" "+QString::number(m_portD));
    m_closeTimer->start(m_dccTimeout*1000);
}

void DccTabItem::slotSomeConnected()
{
    writeEvent(_("Someone connected from %1").arg(m_socket->getRemoteIP()), CONNECT);
    m_closeTimer->stop();
}

//Load lines from logs
void DccTabItem::loadLogLines()
{
    if(!preferences.m_loadLog)
        return;
    clearChat();
    appendIrcStyledText(_("Loading log..."), 11, "", false, false);
    QString logPath = preferences.m_logPath+"/"+getServerName()+"/"+m_nick;
    QApplication::setOverrideCursor(Qt::WaitCursor);
    QDir dir(logPath);
    dir.setFilter(QDir::Files | QDir::Readable);
    dir.setSorting(QDir::Name);
    QStringList files;
    QFileInfoList list = dir.entryInfoList();
    foreach(QFileInfo fileInfo, list)
    {
        if(dxutils::isRightLogFile(fileInfo.absoluteFilePath(), fileInfo.fileName()))
        {
            files << fileInfo.fileName();
        }
    }
    qSort(files.begin(), files.end(), caseInsensitiveLessThan);
    QStringList temp, result;
    for(int i=files.size()-1; i>-1; i--)
    {
        temp.clear();
        std::ifstream file(QString(logPath+"/"+files[i]).toUtf8().constData());
        std::string line;
        while(getline(file, line))
            temp.append(dxutils::normalizeLogLine(files[i], line));
        file.close();
        int nlines = preferences.m_loadLines-result.size();
        if(nlines>0)
        {
            for(int j=temp.size()-1; j>-1&&nlines>0; j--, nlines--)
                result.prepend(temp[j]);
        }
        else break;
    }
    clearChat();
    foreach(QString line, result)
    {
        m_text->append("<div>"+dxutils::parseLogline(line, true)+"</div>");
    }
    QApplication::restoreOverrideCursor();
}

//List log files for tab
void DccTabItem::listLogFiles()
{
    QString logPath = preferences.m_logPath+"/"+m_engine->getServerName()+"/"+m_nick;
    QApplication::setOverrideCursor(Qt::WaitCursor);
    QDir dir(logPath);
    dir.setFilter(QDir::Files | QDir::Readable);
    dir.setSorting(QDir::Name);
    QStringList files;
    QFileInfoList list = dir.entryInfoList();
    foreach(QFileInfo fileInfo, list)
    {
        if(dxutils::isRightLogFile(fileInfo.absoluteFilePath(), fileInfo.fileName()))
        {
            files << fileInfo.fileName();
        }
    }
    qSort(files.begin(), files.end(), caseInsensitiveLessThan);
    QApplication::restoreOverrideCursor();
    if(!files.size())
    {
        appendIrcStyledText(_("No log files for %1-%2.").arg(m_engine->getServerName()).arg(m_nick), 4, "", false, false);
        return;
    }
    appendIrcStyledText(_("Log files:"), 7, "", false, false);
    QString lastYear = dxutils::getParam(files[0], 1, false, '-');
    int lastMonth = dxutils::getParam(files[0], 2, false, '-').toInt();
    appendIrcStyledText(QString("%1:").arg(lastYear), 5, "", false, false);
    appendIrcStyledText(QString("%1:").arg(dxutils::getMonth(lastMonth)), 5, "", false, false);
    QString items = "";
    for(int i=0; i<files.size(); i++)
    {
        if(lastYear!=dxutils::getParam(files[i], 1, false, '-'))
        {
            lastYear = dxutils::getParam(files[i], 1, false, '-');
            lastMonth = dxutils::getParam(files[i], 2, false, '-').toInt();
            appendIrcText(items.left(items.lastIndexOf(',')), "", false, false);
            appendIrcStyledText(QString("%1:").arg(lastYear), 5, "", false, false);
            appendIrcStyledText(QString("%1:").arg(dxutils::getMonth(lastMonth)), 5, "", false, false);
            items.clear();
            items.append(dxutils::getParam(files[i], 3, false, '-')+", ");
        }
        else if(lastMonth!=dxutils::getParam(files[i], 2, false, '-').toInt())
        {
            lastMonth = dxutils::getParam(files[i], 2, false, '-').toInt();
            appendIrcText(items.left(items.lastIndexOf(',')), "", false, false);
            appendIrcStyledText(QString("%1:").arg(dxutils::getMonth(lastMonth)), 5, "", false, false);
            items.clear();
            items.append(dxutils::getParam(files[i], 3, false, '-')+", ");
        }
        else items.append(dxutils::getParam(files[i], 3, false, '-')+", ");
    }
    if(!items.isEmpty()) appendIrcText(items.left(items.lastIndexOf(',')), "", false, false);
}

void DccTabItem::completeHelpCommand()
{
    QString line = m_commandline->text();
    for(int i = 0; i < dxUtils.commandsNo(); i++)
    {
        if(line.section(' ',1).compare(dxUtils.commandsAt(i), Qt::CaseInsensitive) == 0)
        {
            if((i+1) < dxUtils.commandsNo()) m_commandline->setText("/help "+dxUtils.commandsAt(++i));
            else m_commandline->setText("/help "+dxUtils.commandsAt(0));
            break;
        }
        else if(line.section(' ',1).compare(dxUtils.commandsAt(i).left(line.section(' ',1).length()), Qt::CaseInsensitive) == 0)
        {
            m_commandline->setText("/help "+dxUtils.commandsAt(i));
            break;
        }
    }
}

void DccTabItem::writeEvent(const QString &msg, Event event)
{
    switch(event) {
        case DCCMSG:
        {
            bool needHighlight = false;
            if(msg.contains(m_mynick))
                needHighlight = highlightNeeded(msg);
            if(needHighlight) appendIrcStyledText(m_nick+": "+msg, 8, "");
            else appendIrcNickText(m_nick, msg, "");
            if(needHighlight) emit highlight(DCC);
            else emit newmsg(DCC);
        }break;
        case DCCACTION:
        {
            appendIrcStyledText(m_nick+" "+msg, 2, "");
            if(msg.contains(m_mynick) && highlightNeeded(msg)) emit highlight(DCC);
            else emit newmsg(DCC);
        }break;
        case CONNECT:
        {
            appendIrcStyledText(msg, 3, "", true, false);
        }break;
        case DISCONNECT:
        case DCCERROR:
        {
            appendIrcStyledText(msg, 4, "", true, false);
            emit newmsg(DCC);
        }break;
        default:
            break;
    }
}

bool DccTabItem::eventFilter(QObject *obj, QEvent *ev)
{
    if(obj == m_commandline && ev->type() == QEvent::KeyPress)
    {
        QKeyEvent *e = (QKeyEvent *)ev;
        if(e->key() == Qt::Key_Tab)
        {
            QString line = m_commandline->text();
            if(QString::compare(line.section(' ',0,0), "/help", Qt::CaseInsensitive) == 0 && !line.section(' ',1).isEmpty())
            {
                completeHelpCommand();
                return true;
            }
            if(line[0].toLatin1() == '/' && line.section(' ',1).isEmpty())
            {
                for(int i = 0; i < dxUtils.dcccommandsNo(); i++)
                {
                    if(QString::compare(line.section('/',1).section(' ',0,0), dxUtils.dcccommandsAt(i), Qt::CaseInsensitive) == 0)
                    {
                        if((i+1) < dxUtils.dcccommandsNo()) m_commandline->setText("/"+dxUtils.dcccommandsAt(++i)+" ");
                        else m_commandline->setText("/"+dxUtils.dcccommandsAt(0)+" ");
                        break;
                    }
                    else if(QString::compare(line.section('/',1), dxUtils.dcccommandsAt(i).left(line.section('/',1).length()), Qt::CaseInsensitive) == 0)
                    {
                        m_commandline->setText("/"+dxUtils.dcccommandsAt(i)+" ");
                        break;
                    }
                }
                return true;
            }
            return true;
        }
        else if(e->key() == Qt::Key_Up)
        {
            QString line = m_commandline->text();
            if(m_currentPosition!=-1 && m_currentPosition<m_commandsHistory.size())
            {
                if(!line.isEmpty() && line!=m_commandsHistory[m_currentPosition])
                {
                    m_commandsHistory.append(line);
                    if(m_commandsHistory.size() > m_historyMax)
                        m_commandsHistory.removeFirst();
                    m_currentPosition = m_commandsHistory.size()-1;
                }
                if(m_currentPosition > 0 && !line.isEmpty())
                    --m_currentPosition;
                m_commandline->setText(m_commandsHistory[m_currentPosition]);
            }
        }
        else if(e->key() == Qt::Key_Down)
        {
            QString line = m_commandline->text();
            if(m_currentPosition!=-1 && m_currentPosition<m_commandsHistory.size())
            {
                if(!line.isEmpty() && line!=m_commandsHistory[m_currentPosition])
                {
                    m_commandsHistory.append(line);
                    if(m_commandsHistory.size() > m_historyMax)
                        m_commandsHistory.removeFirst();
                    m_currentPosition = m_commandsHistory.size()-1;
                }
                if(m_currentPosition < m_commandsHistory.size()-1)
                {
                    ++m_currentPosition;
                    m_commandline->setText(m_commandsHistory[m_currentPosition]);
                }
                else
                    m_commandline->setText("");
            }
        }
    }
    return QWidget::eventFilter(obj, ev);
}
