/*
 *      ircitem.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "ircitem.h"
#include <QImage>
#include <QFile>
#include "dxutils.h"
#include "preferences.h"
#include "config.h"
#include "i18n.h"

#define MAX_STATUS_SIZE     16
#define SPACING             4

IrcItem::IrcItem(const QString &nick, UserMode mode)
{
    setText(nick);
    setToolTip(nick);
    setMode(mode);
}

void IrcItem::setNick(const QString &nick)
{
    setText(nick);
}

QString IrcItem::getNick()
{
    return text();
}

void IrcItem::setMode(UserMode mode)
{
    setData(mode, IrcItem::RMode);
    updateIconTheme();
}

UserMode IrcItem::getMode()
{
    return static_cast<UserMode>(data(IrcItem::RMode).toInt());
}

void IrcItem::setAway(bool away)
{
    setData(away, IrcItem::RAway);
}

bool IrcItem::getAway()
{
    return data(IrcItem::RAway).toBool();
}

void IrcItem::updateIconTheme()
{
    if(preferences.m_themePath != "internal")
    {
        setIcon(QIcon(dxutils::modeToIcon(getMode())));
    }
}

void IrcItem::updateTooltip(NickInfo nick)
{
    QString tip = QString("%1\n").arg(getNick());
    tip.append(_("User: %1@%2\n").arg(nick.user).arg(nick.host));
    tip.append(_("Realname: %1").arg(nick.real));
    setToolTip(tip);
}

IrcItemDelegate::IrcItemDelegate()
{
}

QSize IrcItemDelegate::sizeHint(const QStyleOptionViewItem& option, const QModelIndex & index) const
{
    QVariant value = index.data(Qt::SizeHintRole);
    if(value.isValid())
        return value.toSize();
    int height = option.fontMetrics.height()+4;
    if(preferences.m_themePath != "internal")
    {
        height = qMax(height, MAX_STATUS_SIZE+2);
    }
    return QSize(80, height);
}

void IrcItemDelegate::paint(QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index) const
{
    painter->save();
    painter->setRenderHint(QPainter::TextAntialiasing);
    QVariant value = index.data(Qt::DecorationRole);
    QColor selectedBg(60, 140, 222);
    QColor selectedText(Qt::white);
    QColor nickTextColor(preferences.m_colors.text);
    QColor awayTextColor = dxutils::makeAwayColor(nickTextColor);
    QFont font;
    if(preferences.m_sameList) font = preferences.m_ircFont;
    QFontMetrics fontMetrics(font);
    bool away = index.data(IrcItem::RAway).toBool();
    QPen penDivision;
    if(option.state & QStyle::State_Selected)
    {
        painter->fillRect(option.rect, selectedBg);
        penDivision.setColor(selectedBg);
        nickTextColor = selectedText;
    }
    QRect rect = option.rect;
    if(preferences.m_themePath != "internal")
    {
        QPixmap pixmap;
        if(value.type() == QVariant::Icon)
        {
            QIcon icon = qvariant_cast<QIcon>(value);
            pixmap = icon.pixmap(QSize(MAX_STATUS_SIZE, MAX_STATUS_SIZE), QIcon::Normal, QIcon::On);
        }
        rect.setWidth(MAX_STATUS_SIZE);
        rect.setHeight(MAX_STATUS_SIZE);
        rect.moveTop(rect.y() + (option.rect.height() - MAX_STATUS_SIZE)/2);
        rect.moveLeft(rect.left() + SPACING);
        painter->drawPixmap(rect, pixmap);
        if(away && QFile::exists(preferences.m_themePath+"/irc_away.png"))
        {
            pixmap = QPixmap(preferences.m_themePath+"/irc_away.png");
            painter->drawPixmap(rect, pixmap);
        }
        rect = option.rect;
        rect.setLeft(rect.x() + MAX_STATUS_SIZE + 2*SPACING);
    }
    else
    {
        UserMode umode = static_cast<UserMode>(index.data(IrcItem::RMode).toInt());
        QString mode = "";
        switch(umode){
            case ADMIN: mode = "!";break;
            case OWNER: mode = "*";break;
            case OP: mode = "@";break;
            case HALFOP: mode = "%";break;
            case VOICE: mode= "+";break;
            case NONE: mode = " ";break;
        }
        int width = fontMetrics.width("@ ");
        rect = option.rect;
        rect.moveTop(rect.y() + (option.rect.height() - fontMetrics.height())/2);
        rect.moveLeft(rect.left() + SPACING);
        painter->setFont(font);
        painter->setPen(nickTextColor);
        painter->drawText(rect, mode);
        rect = option.rect;
        rect.setLeft(rect.x() + width + 2*SPACING);
        rect.moveTop(rect.y() + (option.rect.height() - fontMetrics.height())/2);
    }
    painter->setFont(font);
    if(away)
        painter->setPen(awayTextColor);
    else
        painter->setPen(nickTextColor);
    painter->drawText(rect, index.data(Qt::DisplayRole).toString());
    painter->restore();
}
