/*
 *      tetristabitem.h
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *      Copyright (C) 2007-2008 Graeme Gott <graeme@gottcode.org>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifndef TETRISTABITEM_H
#define TETRISTABITEM_H

#include <QtGui>
#if(QT_VERSION > 0x050000)
#include <QtWidgets>
#endif
#include "defs.h"
#include "tabwidget.h"

static const int columns = 12;
static const int rows = 20;

struct Cell
{
    Cell(int x1 = -1, int y1 = -1)
    :   x(x1),
        y(y1)
    {
    }

    bool operator==(const Cell& cell) const
    {
        return x == cell.x && y == cell.y;
    }

    int x;
    int y;
};

class Game;

class Piece
{
public:
    Piece(int type, Game *parent);

    bool isValid() const { return m_valid; }
    bool moveLeft() { return move(-1, 0); }
    bool moveRight() { return move(1, 0); }
    bool moveDown() { return move(0, 1); }
    bool rotate();
    void drop();
    static void cells(Cell* cells, int type);

private:
    int m_type;
    Game *m_parent;
    Cell m_cells[4];
    Cell m_pivot;
    bool m_valid;

    bool move(int x, int y);
    bool updatePosition(const Cell* ucells);
};

class NextPiece : public QWidget
{
    Q_OBJECT
public:
    NextPiece(QWidget *parent = 0);
public slots:
    void slotPiece(int piece);
protected:
    void paintEvent(QPaintEvent*);
    void resizeEvent(QResizeEvent* event);

    int m_piece;
    int m_anext;
};

class Game : public QWidget
{
    Q_OBJECT
public:
    Game(QWidget *parent = 0);

    bool cell(int x, int y) const;
    void addCell(int x, int y, int type);
    void removeCell(int x, int y, bool update=true);
    void findFullLines();
    void newGame();
    void stopGame();
    void pauseResumeGame();
    void redraw();
    void moveLeft();
    void moveRight();
    void rotate();
    void drop();
    bool isPauseEnable() { return m_pauseEnable; }
    bool isPaused() { return m_paused; }
public slots:
    void slotTimeout();
    void slotNewGame();
    void slotPauseGame();
signals:
    void nextPiece(int piece);
    void updateLabels(int level, int score, int removedLines);
    void started();
    void pausedResumed();
    void finished();
protected:
    virtual void keyPressEvent(QKeyEvent* event);
    void mousePressEvent(QMouseEvent* event);
    void paintEvent(QPaintEvent*);
    void focusOutEvent(QFocusEvent* event);
    void resizeEvent(QResizeEvent* event);
private:
    QTimer *m_timer;
    int m_cells[columns][rows];
    int m_fullLines[4];
    int m_apiece;
    int m_removedLines, m_level, m_score, m_nextPiece;
    bool m_paused, m_done, m_pauseEnable;
    Piece *m_piece;

    void createPiece();
    void removeLines();
    void landPiece();
    void gameOver();
    void updateCell(int x, int y);
};

class TetrisTabItem : public QWidget
{
    Q_OBJECT
public:
    TetrisTabItem(TabWidget *tabbook, int id = 0, QWidget *parent = 0);
    virtual ~TetrisTabItem();

    void setGameFocus();
    int getID() { return m_id; }
public slots:
    void slotNewGame();
    void slotPauseGame();
    void slotGameOver();
    void slotUpdateLabels(int level, int score, int removedLines);
private:
    TabWidget *m_parent;
    QSplitter *m_splitter;
    Game *m_gamecanvas;
    NextPiece *m_nextcanvas;
    QLabel *m_levelLabel, *m_scoreLabel, *m_linesLabel;
    QPushButton *m_newButton, *m_pauseButton;
    int m_id;
};

#endif  /* TETRISTABITEM_H */

